#!/usr/bin/env python3
# 2021 Leo Kaindl

"""
Visualizes tree computed by RAxML-NG using toytree and toyplot.
Renders html using jinja2 and CGI. Visualizes MSA using MView.
Contains both non-primary entry points; -viz and -view.
Allows tree modifications.
"""

import copy
import logging
import math
import multiprocessing
import os
import pickle
import random
import shutil
import socket
import stat
import subprocess
import sys
import webbrowser
import xml.etree.ElementTree as ET
from os import path
from pathlib import Path
from time import time

import numpy as np
import pandas as pd
import toyplot
import toytree
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from jinja2 import Template
from lxml import etree
from toyplot import html, pdf, png, svg, color, data, locator

from ab12phylo_cmd import cli
from ab12phylo_cmd.__init__ import __version__, __author__

# adapted kxlin colors
kxlin = [(.56, 1, .14, 1), (.16, .44, .8, 1), (.06, 1, .75, 1), (.92, 1, .7, 1),
         (.94, .94, .94, .6), (1, 1, 1, 0), (1, 1, 1, 0), (1, 1, 1, 0)] + [(1, 0, 0, 1)] * 20
kxlin_pal = color.Palette(colors=[color.rgba(i[0], i[1], i[2], i[3]) for i in kxlin])
# author colors
pal2 = [(1.0000, 0.9453, 0.9101), (1.0000, 0.9257, 0.1406), (0.9687, 0.7968, 0.6523),
        (0.5117, 0.4648, 0.6132), (0.1640, 0.6757, 0.9882), (0.3750, 0.4726, 0.7851),
        (0.1171, 0.1718, 0.3242), (0.4960, 0.1484, 0.3281), (1.0000, 0.0039, 0.3007),
        (1.0000, 0.4687, 0.6640)]
# pal2 = toyplot.color.Palette(colors=[toyplot.color.rgba(i[0], i[1], i[2], 1) for i in pal2])

blue = (0.1960, 0.5333, 0.7411)
red = (0.6196, 0.0039, 0.2588)

# seaborn rocket palette
rocket = [blue] + [(0.1237, 0.0717, 0.1822), (0.2452, 0.1049, 0.2639),
                   (0.3809, 0.1206, 0.3250), (0.5172, 0.1179, 0.3545),
                   (0.6582, 0.0955, 0.3536), (0.7965, 0.1050, 0.3106),
                   (0.8949, 0.2178, 0.2531), (0.9429, 0.3754, 0.2636),
                   (0.9592, 0.5330, 0.3748), (0.9644, 0.6702, 0.5150),
                   (0.9689, 0.7980, 0.6851)]
# rocket = color.Palette(colors=[color.rgba(t[0], t[1], t[2], 1) for t in rocket])

# seaborn cubehelix palette
cubehelix = [blue] + [[0.1725, 0.1195, 0.2432], [0.2109, 0.1988, 0.3548], [0.2323, 0.2908, 0.4444],
                      [0.2499, 0.3901, 0.5053], [0.2775, 0.4896, 0.5382], [0.3256, 0.5824, 0.5512],
                      [0.3949, 0.6591, 0.5567], [0.4926, 0.7267, 0.5693], [0.6081, 0.7816, 0.6017],
                      [0.7294, 0.8282, 0.6624], [0.8423, 0.8737, 0.7524]]
# cubehelix = color.Palette(colors=[color.rgba(t[0], t[1], t[2], 1) for t in cubehelix])

# blue = color.rgb(blue[0], blue[1], blue[2])
# red = color.rgb(red[0], red[1], red[2])

pg = color.brewer.palette('PinkGreen', reverse=True) \
     + color.Palette(colors=[color.rgb(.26, .26, .26), color.brewer.palette('Spectral')[1]])

base_legend = '<svg class="toyplot-canvas-Canvas" xmlns:toyplot="http://www.sandia.gov/toyplot" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns="http://www.w3.org/2000/svg" width="100.0px" height="40.0px" viewBox="0 0 100.0 40.0" preserveAspectRatio="xMidYMid meet" style="background-color:transparent;border-color:#292724;border-style:none;border-width:1.0;fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:Helvetica;font-size:12px;opacity:1.0;stroke:rgb(16.1%,15.3%,14.1%);stroke-opacity:1.0;stroke-width:1.0" id="ta084f8acf92540e38738f50593e76861"><g class="toyplot-coordinates-Table" id="t9e2f7ff350ea4dbebe0c12a7395bb879"><rect x="0.0" y="0.0" width="20.0" height="20.0" style="fill:rgb(56%,100%,14%);fill-opacity:1.0;stroke:rgb(100%,100%,100%);stroke-opacity:1.0;stroke-width:4" /><rect x="20.0" y="0.0" width="20.0" height="20.0" style="fill:rgb(16%,44%,80%);fill-opacity:1.0;stroke:rgb(100%,100%,100%);stroke-opacity:1.0;stroke-width:4" /><rect x="40.0" y="0.0" width="20.0" height="20.0" style="fill:rgb(6%,100%,75%);fill-opacity:1.0;stroke:rgb(100%,100%,100%);stroke-opacity:1.0;stroke-width:4" /><rect x="60.0" y="0.0" width="20.0" height="20.0" style="fill:rgb(92%,100%,70%);fill-opacity:1.0;stroke:rgb(100%,100%,100%);stroke-opacity:1.0;stroke-width:4" /><rect x="80.0" y="0.0" width="20.0" height="20.0" style="fill:rgb(94%,94%,94%);fill-opacity:0.6;stroke:rgb(100%,100%,100%);stroke-opacity:1.0;stroke-width:4" /><g transform="translate(10.0,30.0)"><text x="-4.332" y="3.066" style="fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:helvetica;font-size:12.0px;font-weight:bold;stroke:none;vertical-align:baseline;white-space:pre">A</text></g><g transform="translate(30.0,30.0)"><text x="-4.332" y="3.066" style="fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:helvetica;font-size:12.0px;font-weight:bold;stroke:none;vertical-align:baseline;white-space:pre">C</text></g><g transform="translate(50.0,30.0)"><text x="-4.668" y="3.066" style="fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:helvetica;font-size:12.0px;font-weight:bold;stroke:none;vertical-align:baseline;white-space:pre">G</text></g><g transform="translate(70.0,30.0)"><text x="-3.666" y="3.066" style="fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:helvetica;font-size:12.0px;font-weight:bold;stroke:none;vertical-align:baseline;white-space:pre">T</text></g><g transform="translate(90.0,30.0)"><text x="-4.332" y="3.066" style="fill:rgb(16.1%,15.3%,14.1%);fill-opacity:1.0;font-family:helvetica;font-size:12.0px;font-weight:bold;stroke:none;vertical-align:baseline;white-space:pre">N</text></g></g></svg>'


def _visualize(*args):
    """Entry point for re-visualization"""
    # if no argument -> default to cwd
    args = sys.argv[1:]
    if len(args) == 0:
        args = [os.getcwd()]
    args = ['--visualize'] + args
    # parse to get defaults
    namespace = cli.parser(args, visualize=True).args
    tree_build(namespace)
    print('BYE!')


def _view():
    """Entry point for re-view"""
    # if no argument -> default to cwd
    args = sys.argv[1:]
    if len(args) == 0:
        args = [os.getcwd()]
    args = ['-view'] + args
    # parse to get defaults
    namespace = cli.parser(args, view=True).args
    tree_view(namespace.dir)
    print('BYE!')


def tree_view(_dir):
    """
    Opens a result page and starts a CGI server
    :param _dir:
    :return:
    """
    log = logging.getLogger(__name__)

    # open in new tab
    webbrowser.open(path.join(_dir, 'result.html'), new=2)

    # find original port
    port = int(etree.parse(path.join(_dir, 'result.html'), parser=etree.HTMLParser())
               .findall('//meta[@name="port"]')[0].get('content'))

    # start CGI server
    try:
        py3 = sys.executable
        log.debug('python3 executable at %s' % py3)
        log.info('starting CGI server on port: %d for 20min' % port)
        subprocess.run('%s -m http.server --cgi %d' % (py3, port), shell=True,
                       stdout=subprocess.PIPE, cwd=_dir, timeout=1200)
    except subprocess.TimeoutExpired:
        txt = 'CGI server shut down after timeout. If necessary, re-start in %s via ' \
              '"%s -m http.server --cgi %d" or re-run ab12phylo-view' % (_dir, sys.executable, port)
        log.info(txt)
        print(txt, file=sys.stderr)
    except KeyboardInterrupt:
        txt = 'AB12PHYLO shut down.'
        log.info(txt)
        print('\n' + txt, file=sys.stderr)
    return


def mview_msa(args):
    """
    Using MView, create an .html visualization for the annotated MSA.
    """
    # hacky re-configuration
    mv_path = path.join(path.abspath(path.dirname(__file__)),
                        'tools', 'mview-1.67').replace(' ', '\ ')
    perl_binary = shutil.which('perl')
    if not perl_binary:
        raise ValueError
    lines = open(path.join(mv_path, 'bin', 'mview'), 'r').read().split('\n')
    # monkey patch
    lines[0] = '#!' + perl_binary
    line = lines[25].split('"')
    line[1] = mv_path
    lines[25] = '"'.join(line)

    # over-write file
    open(path.join(mv_path, 'bin', 'mview'), 'w').write('\n'.join(lines))

    used_msa = args.new_msa if path.isfile(args.new_msa) else args.msa

    arg = '%s %s -in fasta -moltype dna -width 80 -conservation on -coloring consensus ' \
          '-threshold 80 -label2 -label4 -label5 -consensus on -con_threshold 80 ' \
          '-html head -css on  -colormap leo "%s" > "%s"' % \
          (perl_binary, path.join(mv_path, 'bin', 'mview'), used_msa, args.mview_msa)

    p = subprocess.run(arg, shell=True, stdout=subprocess.PIPE)
    return 'MView run ' + p.stdout.decode('utf-8').strip()


class tree_build:
    """
    Tree visualization based on toytree and toyplot. Also calls tree_view()
    """

    def __init__(self, _args):
        self.log = logging.getLogger(__name__)
        self.args = _args
        self.replace = _args.replace

        self.args.threshold *= 10
        self.tree_file = _args.final_tree + '_%s.nwk' % _args.metric

        self.log.debug('toyplot %s' % toyplot.__version__)
        self.log.debug('toytree %s' % toytree.__version__)
        self.log.debug(sys.version_info)

        preview = self._preview_topology()

        render_dict = self._annotate_msa()

        self._edit1()

        self.log.debug('drawing circular tree')
        start = time()
        try:
            ctup = self.tree.draw(width=800, height=800, scalebar=True,
                                  node_sizes=list(self.tree.get_node_values('size', 1, 1)),
                                  node_colors=[color.rgb(n[0], n[1], n[2]) for n in list(
                                      self.tree.get_node_values('color', 1, 1))],
                                  tip_labels=True, tip_labels_align=False,
                                  tip_labels_colors=[color.rgb(rocket[n][0], rocket[n][1], rocket[n][2])
                                                     for n in list(self.tree.get_node_values('score', 1, 1))
                                                     if n != -1][::-1],
                                  layout='c')
            ctup[0].style['background-color'] = 'white'
            ctup[1].show = False
            if not self.args.out_fmt:
                self.args.out_fmt = ['pdf']

            if 'png' in self.args.out_fmt:
                png.render(ctup[0], path.join(self.args.dir, 'circular.png'), scale=1.6)
            if 'pdf' in self.args.out_fmt or len(self.args.out_fmt) == 0:
                pdf.render(ctup[0], path.join(self.args.dir, 'circular.pdf'))
            if 'svg' in self.args.out_fmt:
                svg.render(ctup[0], path.join(self.args.dir, 'circular.svg'))
            self.log.info('rendered circular in %.2f sec' % (time() - start))

        except ET.ParseError as ex:
            self.log.error('XML ParseError. Invalid character in a sample ID? Please check metadata.tsv')
            sys.exit(1)
        except ValueError as ex:
            self.log.exception(ex)

        # prep rectangular tree: save species and pid to name
        tree_no_msa = copy.deepcopy(self.tree)

        # save species and pid to name
        for node in tree_no_msa.treenode.traverse():
            if node.is_leaf():
                try:
                    if node.name in node.species:
                        node.name = node.species
                    else:
                        node.name += ' ' + node.species
                        node.name += ' %.2f' % node.pid
                except AttributeError:
                    continue

        # write new-ish newick file
        tree_no_msa.write(self.args.annotated_tree, tree_format=0)

        self.log.debug('drawing tree without msa')
        start = time()
        try:
            # get dim for canvas
            w, h = 1200, len(self.tips) * 14 + 80
            if self.args.print_supports:
                rtup = tree_no_msa.draw(width=w, height=h, scalebar=True, tip_labels=True, node_labels='support',
                                        node_labels_style={'font-size': '6px', 'fill': '#FFFFFF',
                                                           'baseline-shift': '-1px',
                                                           'font-weight': 'bold'},
                                        node_sizes=list(self.tree.get_node_values('size', 1, 1)),
                                        node_colors=[color.rgb(n[0], n[1], n[2]) for n in
                                                     list(self.tree.get_node_values('color', 1, 1))],
                                        tip_labels_colors=[color.rgb(rocket[n][0], rocket[n][1], rocket[n][2])
                                                           for n in list(self.tree.get_node_values('score', 1, 1))
                                                           if n != -1][::-1])
            else:
                rtup = tree_no_msa.draw(width=w, height=h, scalebar=True, tip_labels=True,
                                        node_sizes=list(self.tree.get_node_values('size', 1, 1)),
                                        node_colors=[color.rgb(n[0], n[1], n[2]) for n in
                                                     list(self.tree.get_node_values('color', 1, 1))],
                                        tip_labels_colors=[color.rgb(rocket[n][0], rocket[n][1], rocket[n][2])
                                                           for n in list(self.tree.get_node_values('score', 1, 1))
                                                           if n != -1][::-1])
            rtup[0].style['background-color'] = 'white'
            rtup[1].y.show = False
            rtup[1].x.show = True
            rtup[1].x.domain.max = self.tree.treenode.height / 5  # 0 is right-most tip of tree. -> divide space!

            if 'png' in self.args.out_fmt:
                png.render(rtup[0], path.join(self.args.dir, 'rectangular.png'), scale=1.6)
            if 'pdf' in self.args.out_fmt:
                pdf.render(rtup[0], path.join(self.args.dir, 'rectangular.pdf'))
            if 'svg' in self.args.out_fmt:
                svg.render(rtup[0], path.join(self.args.dir, 'rectangular.svg'))
            self.log.info('rendered rectangular in %.2f sec' % (time() - start))

        except ValueError as ex:
            self.log.exception(ex)

        try:
            self.log.info(mview_msa(self.args))
        except Exception as ex:
            with open(self.args.mview_msa, 'w') as fh:
                fh.write('MView (Perl) failed\n')
                fh.write(str(ex))

        # render rectangular tree with MSA
        reserve_gap = False
        if type(self.args.msa_viz) == list and 'png' in self.args.msa_viz:
            try:
                reserve_gap = True
                proc = multiprocessing.Process(target=self._with_matrix,
                                               args=([kxlin_pal]))
                proc.start()
            except KeyboardInterrupt:
                self.log.warning('cancel msa_viz')

        # but display it any way if it exists
        if reserve_gap or path.isfile(path.join(self.args.dir, 'rectangular_msa.png')):
            render_dict['msa_viz'] = '<div id="msa_viz" href="msa_viz" class="section level3">' \
                                     '<h3><strong>3.2</strong> +MSA</h3>' \
                                     '<img id="msa_viz" src="rectangular_msa.png" alt="tree+MSA" ' \
                                     'style="width:860px"/><h4>%s</h4></div>' % base_legend
        else:
            render_dict['msa_viz'] = ''

        self._write_html(render_dict, preview, ctup[0], rtup[0], tree_no_msa)
        if not self.args.headless:
            tree_view(self.args.dir)

    def _preview_topology(self):
        """
        Reads in the tree file, and saves a .png with rectangular and circular topology previews.

        :return: toyplot.Canvas object
        """
        self.log.debug('reading tree')
        # read in the tree
        self.tree = toytree.tree(open(self.tree_file, 'r').read(), tree_format=0)

        # drop / replace nodes
        if self.args.replace_nodes:
            for idx in self.args.replace_nodes:
                self.tree.idx_dict[idx].add_sister(name='%d_replaced' % idx, dist=1)
                self.tree.idx_dict[idx].detach()
            # saving the node names now saves some work
            self.args.replace_nodes = ['%d_replaced' % idx for idx in self.args.replace_nodes]
        else:
            # save empty list
            self.args.replace_nodes = []
        if self.args.drop_nodes:
            [self.tree.idx_dict[idx].detach() for idx in self.args.drop_nodes]

        # outgroup rooting
        if self.args.root:
            self.tree = self.tree.root(names=[self.tree.idx_dict[self.args.root].name])

        self.tree = toytree.tree(self.tree.write())

        # set dimensions of the canvas
        preview = toyplot.Canvas(width=800, height=400, style={'background-color': 'white'})
        # dissect canvas into two cartesian areas
        ax0 = preview.cartesian(bounds=('5%', '48%', '5%', '95%'))
        ax1 = preview.cartesian(bounds=('52%', '95%', '5%', '95%'))

        # call draw with the 'axes' argument to pass it to a specific cartesian area
        self.log.debug('drawing preview')
        self.tree.draw(axes=ax0, layout='r', tip_labels=False)
        self.tree.draw(axes=ax1, layout='c', tip_labels=False)

        # hide the axes coordinates
        ax0.show = False
        ax1.show = False

        # render to image
        png.render(preview, self.args.topo)
        self.log.info('rendered preview %s' % self.args.topo)
        return preview

    def _annotate_msa(self):
        """
        Reads genes and no_BLAST from log. Reads in metadata for first gene.
        Writes re-named, re-ordered msa and saves records if necessary.

        :return some information for HTML rendering as dict
        """
        if self.args.finish:
            try:
                log = open(self.args.log[:-5] + '1.log', 'r').read()
                self.args.log = self.args.log[:-5] + '1.log'
            except FileNotFoundError:
                try:
                    log = open(self.args.log[:-7] + '.log', 'r').read()
                    self.args.log = self.args.log[:-7] + '.log'
                except FileNotFoundError:
                    self.log.error('no log file found')
                    exit(1)
        else:
            try:
                log = open(self.args.log, 'r').read()
            except FileNotFoundError:
                log = open(self.args.log[:-4] + '-p1.log', 'r').read()
                self.args.log = self.args.log[:-4] + '-p1.log'

        # jinja
        render_info = dict()
        render_info['run_start'] = log[:19]
        render_info['run_args'] = log[54:log.find('\n')]

        # extract genes
        start = log.find('--GENES--') + 10
        end = log[start:start + 800].find('\n')
        self.genes = log[start:start + end].split('::')
        self.log.debug('read genes from log: %s' % ':'.join(self.genes))
        gene = self.genes[0]
        self.log.debug('using annotation for %s' % gene)

        # extract seed
        start = log.find('seed for this run:') + 19
        end = log[start:start + 20].find('\n')
        render_info['seed'] = int(log[start:start + end])

        if '--SKIPPING BLAST--' in log or 'BLAST+ not installed' in log:
            self.args.no_BLAST = True

        # fetch tips labels
        self.tips = list(reversed(self.tree.get_tip_labels()))
        # jinja
        render_info['num_seq'] = str(len(self.tips))
        render_info['genes'] = ', '.join(self.genes)

        # read in tabular data
        self.df = pd.read_csv(self.args.tsv, sep='\t', dtype={'id': str})
        self.df.set_index('id', inplace=True)
        # drop and order rows
        self.df = self.df[self.df.gene == gene]
        # add dummy data to data frame
        for node_id in self.args.replace_nodes:
            self.df.at[node_id, 'gene'] = gene
        self.df = self.df.reindex(self.tips)

        # read in MSA
        records = {record.id: record.seq for record in SeqIO.parse(self.args.msa, 'fasta')}

        # write re-named, re-ordered MSA
        with open(self.args.new_msa, 'w') as new_msa:
            for tip in self.tips:
                if tip in self.args.replace_nodes:
                    continue
                # get seq and cut out artificial separator
                seq = str(records[tip]).replace(self.args.sep, '')
                # get metadata
                entry = self.df.loc[tip]

                des = ''
                # get BLAST info
                if 'BLAST_species' in entry and not pd.isnull(entry.BLAST_species):
                    des = entry.BLAST_species
                    # annotate BLAST score
                    pid = '%.2f' % entry.pid if entry.pid < 100 else '100'
                    des += '; pid %s' % pid

                # rename REFs
                if tip.startswith('REF_'):
                    if len(self.genes) > 1 and 'strain' in entry.reference_species:
                        des, tip = entry.reference_species.split(' strain ')
                    else:
                        tip = entry.accession
                        des = entry.reference_species

                # get replaced IDs
                if 'replaces' in entry and not pd.isnull(entry.replaces):
                    des += '; replaces ' + entry.replaces

                SeqIO.write(SeqRecord(Seq(seq), id=tip, description=des), new_msa, 'fasta')
            self.log.debug('wrote updated MSA: %s' % self.args.new_msa)

        # get lengths of genes
        self.g_lens = [(gene, len(seq)) for gene, seq in
                       zip(self.genes, records[next(iter(records))].split(self.args.sep))]

        # save seqs if necessary
        if type(self.args.msa_viz) == list:
            self.records = records
            # make empty dummy seqs
            seq = '-' * [item[1] for item in self.g_lens if item[0] == gene][0]
            for _id in self.args.replace_nodes:
                self.records[_id] = Seq(seq)

        self.log.debug('gene lengths: %s' % self.g_lens)
        return render_info

    def _with_matrix(self, pal):
        """
        Create a toytree with a matrix representation of the MSA.
        """
        start = time()
        self.log.warning('drawing tree with msa. You can interrupt and proceed via Cmd+C.')
        start = time()

        # if no format was specified, render .PNGs
        if not self.args.msa_viz:
            self.args.msa_viz = ['png']

        # get ticks for gene ends
        ticks = [item[1] for item in self.g_lens]
        for i in range(1, len(ticks)):
            ticks[i] += ticks[i - 1] + 10
        # ticks[-1] -= 1
        ticks = [p - 3 for p in ticks]  # shift sideways
        self.log.debug('ticks: %s' % ':'.join(str(tick) for tick in ticks))

        # get seqs as numpy array in plot order
        seqs = np.array([self.records[_id] for _id in self.tips])
        self.log.info('numpy array representation of MSA\n\t'
                      'shape: %s entries: %s' % (seqs.shape, ':'.join(np.unique(seqs))))
        bases = ['A', 'C', 'G', 'T', 'N', '-', ' ', 'S']
        weird_chars = np.setdiff1d(np.unique(seqs), bases)
        self.log.debug('observed: %s' % ':'.join(set(list(weird_chars) + bases)))
        # get a mapping from characters to integer
        codes = dict(zip(list(bases) + list(weird_chars), range(len(bases) + len(weird_chars))))
        # translate seqs array to integers
        coded_seqs = np.vectorize(codes.get)(seqs)

        # get a data table of other info
        cols = set(self.df.columns).intersection({'reference_species', 'BLAST_species'})
        lines = self.df[cols]

        if lines.shape[1] == 2:
            # re-order columns
            lines = lines.reindex(columns=['BLAST_species', 'reference_species'])
            # paste ref species for references
            for j in range(lines.shape[0]):
                if not pd.isna(lines.iloc[j, 1]):
                    lines.iloc[j, 0] = lines.iloc[j, 1]

        # annotate missing species
        if self.args.no_BLAST or 'pid' not in self.df:
            dt = lines.fillna('')
        else:
            lines = lines.fillna('no BLAST hit')
            dt = pd.concat([self.df.loc[:, ['pid']], lines.iloc[:, 0]], axis=1)
        # rename last column
        try:
            dt.columns = list(dt.columns)[:-1] + ['species']
            dt = toyplot.data.Table(dt)
        except ValueError:
            dt['species'] = ''

        hh, ww = coded_seqs.shape

        # create a canvas
        w = int(ww * 2.2 + hh * 2 + 80)
        h = hh * 16 + 80

        self.log.debug('dim w/ msa %spx by %spx' % (w, h))
        rcanvas = toyplot.Canvas(width=w, height=h, style={'background-color': 'white'})
        axes = rcanvas.cartesian(bounds=(40, 0.26 * w, 40, h - 40))  # xstart xend ystart yend

        self.log.debug('drawing tree')
        if self.args.print_supports:
            self.tree.draw(axes=axes, scalebar=True, tip_labels=True, tip_labels_align=True,
                           node_labels='support', node_labels_style={'font-size': '6px',
                                                                     'fill': '#FFFFFF',
                                                                     'baseline-shift': '-1px',
                                                                     'font-weight': 'bold'},
                           edge_align_style={'stroke-width': .7, 'stroke': 'silver'},
                           node_sizes=list(self.tree.get_node_values('size', 1, 1)),
                           node_colors=[color.rgb(n[0], n[1], n[2]) for n in list(
                               self.tree.get_node_values('color', 1, 1))],
                           tip_labels_colors=[color.rgb(n[0], n[1], n[2]) for n in list(
                               self.tree.get_node_values('type', 1, 1)) if n != 0][::-1])
        else:
            self.tree.draw(axes=axes, scalebar=True, tip_labels=True, tip_labels_align=True,
                           edge_align_style={'stroke-width': .7, 'stroke': 'silver'},
                           node_sizes=list(self.tree.get_node_values('size', 1, 1)),
                           node_colors=[color.rgb(n[0], n[1], n[2]) for n in list(
                               self.tree.get_node_values('color', 1, 1))],
                           tip_labels_colors=[color.rgb(n[0], n[1], n[2]) for n in list(
                               self.tree.get_node_values('type', 1, 1)) if n != 0][::-1])
        axes.y.show = False
        axes.x.show = True

        self.log.info('drawing matrix')
        msa = rcanvas.matrix((coded_seqs, color.CategoricalMap(palette=pal)),
                             lshow=False, bshow=True, step=100, bounds=(0.26 * w - 40, w, 0, h),
                             tlocator=locator.Explicit(ticks, [i[0] + ']' for i in self.g_lens],
                                                       format='{:>12}'.format))
        msa.body.gaps.rows[...] = 7

        # # add BLAST score indicators
        # self.log.debug('adding annotations')
        # if not self.args.no_BLAST:
        #     axa = msa.right.column[-2].cartesian()
        #     axa.scatterplot(np.zeros(len(self.tips)) + 8, np.arange(len(self.tips)),
        #                     color=colors[4], size=8, marker='s')
        #     msa.right.column[-2].width = 30
        msa.right.column[-2].width = 12
        msa.left.column[1].width = 44

        # write species column
        msa.right.column[-1].data = dt['species']
        msa.right.column[-1].width = 240
        msa.right.column[-1].align = 'left'
        msa.right.column[-1].lstyle = {'font-weight': 'regular'}

        # colorize species annotation for ref and missing
        shift = msa.shape[0] - len(self.tips) - 2  # line index shift because of 'top'
        for j in range(len(self.tips)):
            msa.cells.cell[j + shift, -1].lstyle = \
                {'fill': [color.rgb(cubehelix[n][0], cubehelix[n][1], cubehelix[n][2])
                          for n in list(self.tree.get_node_values('score', 1, 1))
                          if n != -1][j]}
        self.log.info('rectangular tree setup in %.2f sec' % (time() - start))

        if 'png' in self.args.msa_viz:
            png.render(rcanvas, path.join(self.args.dir, 'rectangular_msa.png'), scale=2)
        if 'pdf' in self.args.msa_viz:
            pdf.render(rcanvas, path.join(self.args.dir, 'rectangular_msa.pdf'))
        self.log.info('rendered with msa in %.2f sec' % (time() - start))

    def _edit1(self):
        """
        type -> reference, normal seq or no BLAST hit
        BLAST -> colors change from black to red to glowy yellow in 2% steps
        support -> larger dot means better support.
        :return:
        """
        multi = True if len(self.genes) > 1 else False

        # guess if support val.s are in [0-1] or in [0-100], then scale with 1 or 1/100 resp.
        sup_scaler = 1 if 1.1 > max([float(s) for s in
                                    self.tree.get_node_values('support') if s]) else 100
        # rename reference nodes and add features
        for node in self.tree.treenode.traverse():
            if node.is_leaf():
                entry = self.df.loc[node.name]

                if node.name.startswith('REF_'):
                    if multi and 'strain' in entry.reference_species:
                        entry.at['reference_species'], node.name = entry.reference_species.split(' strain ')
                    else:
                        node.name = entry.accession
                    node.add_feature('type', blue)
                    node.add_feature('species', entry.reference_species)
                    node.add_feature('score', 0)  # this is the position of blue in the palette

                elif 'BLAST_species' in entry and not pd.isnull(entry.BLAST_species):
                    node.add_feature('type', (.2, .2, .2, 1))
                    node.add_feature('species', entry.BLAST_species)
                    # node.add_feature('score', '%.2f' % entry.pid if entry.pid < 100 else '100')
                    node.add_feature('pid', entry.pid)
                    node.add_feature('score', 1 + math.ceil(min(100 - entry.pid, 20) / 2))  # map to the palette

                else:
                    node.add_feature('type', rocket[1] if self.args.no_BLAST else red)
                    node.add_feature('score', 1 if self.args.no_BLAST else 11)
                    # dark gray if no BLASTing, else bad light color

                if self.replace and 'replaces' in entry and not pd.isnull(entry.replaces):
                    node.name += ', ' + entry.replaces

                node.add_feature('size', 0)
                node.add_feature('color', kxlin[1])  # just so the field exists
            else:
                # use support values for node size
                min_size = 2
                node.add_feature('size', min_size + float(node.support) / sup_scaler * 10 if node.support else 0)
                node.add_feature('color', kxlin[1] if node.size > self.args.threshold + min_size else pal2[8])
                node.add_feature('score', -1)  # just so the field exists
                node.add_feature('type', 0)  # just so the field exists

            # some tree tweaks
            if self.args.print_supports:
                node.support = int(round(node.support * 100 / sup_scaler))
            if self.args.min_plot_dist:
                node.dist = max(node.dist, self.args.min_plot_dist)
        return

    def _write_html(self, materials, topo, circ, rect, tree):
        """
        Write interactive output .HTML using Jinja2
        :return:
        """
        self.log.debug('creating HTML')

        start = time()
        materials['topo'] = toyplot.html.tostring(topo)
        materials['circular'] = toyplot.html.tostring(circ)
        materials['rectangular'] = toyplot.html.tostring(rect)
        self.log.info('rendered embeddable HTMLs in %.2f sec' % (time() - start))

        start = time()
        materials['missing'] = open(self.args.missing_samples, 'r').read() \
            if os.path.exists(self.args.missing_samples) else 'no missing samples'
        materials['msa'] = open(self.args.new_msa, 'r').read()
        materials['newick'] = open(self.args.annotated_tree, 'r').read()
        # fully insert MView HTML
        # materials['mview'] = open(self.args.mview_msa, 'r').read()
        self.log.debug('read files in %.2f sec' % (time() - start))

        materials['metric'] = self.args.metric
        materials['gap'] = self.args.gap_share
        materials['unknown'] = self.args.unknown_share
        materials['poly'] = self.args.poly_allelic
        materials['g_lens'] = '_'.join([entry[0] + ':' + str(entry[1]) for entry in self.g_lens])
        materials['msa_path'] = path.basename(self.args.msa)  # path.abspath(self.args.new_msa)
        materials['mview'] = path.basename(self.args.mview_msa)
        materials['threshold'] = int(self.args.threshold * 10)
        materials['min_dist'] = self.args.min_dist
        materials['metadata'] = path.relpath(self.args.tsv, self.args.dir)
        materials['version'] = __version__
        materials['author'] = __author__

        # find a free port. susceptible to race conditions
        sock = socket.socket()
        try:
            sock.bind(('localhost', random.randint(8000, 8010)))
        except OSError:
            sock.bind(('localhost', 0))
        port = sock.getsockname()[1]
        sock.close()
        materials['port'] = port

        start = time()
        df = pd.read_csv(self.args.tsv, sep='\t', dtype={'id': str})
        df.set_index('id', inplace=True)
        # convert box to integer
        df.box = df.box.fillna(0).astype(int)
        # replace missing box with -
        df.box = df.box.replace(0, '-')
        # display only two digits
        pd.options.display.float_format = '{:,.2f}'.format
        if 'pid' in df:
            # print 100 as exact number
            df.pid = df.pid.replace(100.00, '100')
        # print only file names, not full path
        df.file = df.file.fillna('')
        df.file = df.file.apply(lambda _path: _path.split('/')[-1])
        materials['table'] = df.to_html(na_rep='', justify='justify')

        df2 = pd.read_csv(self.args.bad_seqs, sep='\t', dtype={'id': str})
        # re-order columns
        df2 = df2.reindex(columns=['id', 'problem', 'gene', 'box', 'file'])
        df.box = df.box.replace(0, '-')
        materials['bad_seqs'] = df2.to_html(na_rep='', index=False, justify='justify')
        self.log.debug('rendered tables in %.2f sec' % (time() - start))

        # dump tree
        with open(path.join(self.args.dir, 'tree.pickle'), 'wb') as out_file:
            pickle.dump(tree, out_file)
        materials['pickle'] = 'tree.pickle'  # path.abspath(path.join(self.args.dir, 'tree.pickle'))

        # copy cgi-bin directory to output directory
        _dst = path.join(self.args.dir, 'cgi-bin')
        # to keep compatibility with python3.6, don't use shutil.copytree(dirs_exist_ok)
        try:
            # delete old directory
            shutil.rmtree(path=_dst)
        except FileNotFoundError:
            pass
        shutil.copytree(src=path.join(path.dirname(__file__), 'cgi-bin'),
                        dst=path.join(self.args.dir, 'cgi-bin'))

        # copy logo and icon
        _img = path.join(self.args.dir, '.img')
        os.makedirs(_img, exist_ok=True)
        shutil.copy(src=path.join(_dst, 'favicon.png'), dst=path.join(_img, 'favicon.png'))
        shutil.copy(src=path.join(_dst, 'favicon.ico'), dst=path.join(_img, 'favicon.ico'))
        self.log.debug('copied files to output directory')

        start = time()
        self.log.debug('rendering output HTML')
        materials['log'] = open(self.args.log, 'r').read()
        template = Template(open(path.join(path.dirname(__file__), 'template.jinja'), 'r').read())
        rendered_html = template.render(materials)

        # make CGI script executable a bit later so it's surely there
        cgi_file = Path(self.args.dir) / 'cgi-bin' / 'ab12phylo.py'
        cgi_file.chmod(cgi_file.stat().st_mode | stat.S_IEXEC | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        open(path.join(self.args.dir, 'result.html'), 'w').write(rendered_html)
        return



# 2021 Leo Kaindl

import logging
import os
import shutil
import stat
import subprocess
import threading
from argparse import Namespace
from pathlib import Path
from time import sleep

import gi
import pandas as pd
from Bio import SeqIO
from numpy import isnan, nan

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ab12phylo_cmd import blast
from ab12phylo import repo
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 5


class blast_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.blast_spinner.stop()
        iface.spinbox.set_visible(False)

        iface.remote_db.set_model(data.remote_dbs)
        iface.remote_db.get_child().connect('focus_out_event', self._save_custom_remote_db)

        # connect buttons
        iface.blast_local.connect('clicked', self.start_BLAST, 'local')
        iface.blast_remote.connect('clicked', self.start_BLAST, 'remote')
        iface.blast_import.connect('clicked', self.start_BLAST, None)
        iface.xml_import.connect('file_set', self.start_BLAST, None)
        iface.blast_exe.connect('file_set', lambda *args: self.prep0(iface.blast_exe.get_filename()))
        iface.blast_db.connect('button-release-event', lambda co, *args: co.
                               popdown() if co.props.popup_shown else co.popup())
        iface.blast_gene.connect('changed', lambda *args: self._refill())

        # set up the species annotation table
        spi = iface.sp_info
        spi.connect_after('size-allocate', lambda *args: [col.
                          set_max_width(spi.get_allocated_width() // 2) for col in spi.get_columns()])
        spi.set_model(data.sp_model)
        spi.set_reorderable(True)
        for i, title in enumerate(['sample', 'pid', 'species', 'runner-up']):
            crt = Gtk.CellRendererText(editable=True)
            crt.connect('edited', self._save_sp_edit, spi, i)
            col = Gtk.TreeViewColumn(title=title, cell_renderer=crt, text=i, foreground=4)
            col.set_sort_column_id(i)
            col.set_resizable(True)
            spi.append_column(col)
        spi.columns_autosize()

        iface.blast_seen = False

    def prep0(self, path):
        data = self.data
        iface = self.iface
        iface.i = 0
        iface.k = 3
        iface.text = 'search for BLAST+ installation'
        if path:
            path = Path(path).resolve()
            if path.is_file():
                path = path.parent  # path is supposed to be a directory!
            data.blast_path = path  # save path to executable in project

        # check if all necessary BLAST+ scripts are available
        try:
            binary = ab12phylo_app_base.CFG.get('blastn', shutil.which('blastn'))
        except KeyError:
            binary = shutil.which('blastn', path=path)

        scripts = ['blastn', 'blastdbcmd', 'update_blastdb.pl', 'makeblastdb']
        if not binary:
            missing = scripts
        else:
            missing = list()
            binary = Path(binary)

            for sc in scripts:
                try:
                    exe = next(binary.parent.rglob(f'{sc}{os.getenv("PATHEXT", default="")}'))
                    # Make the file executable
                    exe.chmod(exe.stat().st_mode | stat.S_IEXEC)
                except StopIteration:
                    missing.append(sc)

        if not missing:
            data.blast_path = binary.parent
        else:
            data.blast_path = None
            self.show_notification(msg='Necessary scripts could not be '
                                       'found on the $PATH:', items=missing)
            sleep(.1)
            GObject.idle_add(self.prep4)
            return True

        iface.blast_exe.set_filename(str(data.blast_path))
        mo = Gtk.ListStore(str, str, str, str, bool)
        iface.blast_db.set_model(mo)  # shared model
        iface.db_info.set_model(mo)
        iface.db_info.set_reorderable(True)
        for i, title in enumerate(['database', 'size', 'title']):
            col = Gtk.TreeViewColumn(title=title, cell_renderer=Gtk.CellRendererText(),
                                     text=i + 1, underline=4)
            col.set_resizable(True)
            iface.db_info.append_column(col)

        iface.thread = threading.Thread(target=self.prep1, args=[data.blast_path, mo])
        GObject.timeout_add(50, self.update, PAGE)
        iface.i = 1
        iface.thread.start()
        return True

    def prep1(self, path, mo):
        """
        There is a BLAST+ executable on the $PATH. Pre-select it, search for local and remote
        databases with BLAST+ commands, then set-up the db_info and the blast_db ComboBox.
        :param gui:
        :param binary:
        :return:
        """
        data = self.data
        iface = self.iface
        try:
            iface.text = 'look for local databases'
            output = subprocess.check_output(
                str(path / 'blastdbcmd') + ' -recursive -list_outfmt "%f\t%t\t%U" -list ' +
                str(repo.BASE_DIR), shell=True).decode('utf-8').strip().split('\n')
            if output and output != ['']:
                sleep(.05)
                GObject.idle_add(self.prep2, iface.db_info, output, mo)
                iface.i = 2
        except subprocess.CalledProcessError as ex:
            LOG.error(ex)

        try:
            iface.text = 'look for online pre-compiled databases'
            output = subprocess.check_output(str(path / 'update_blastdb.pl') +
                                             ' --source gcp --showall tsv',
                                             shell=True).decode('utf-8')
            sleep(.05)
            GObject.idle_add(self.prep3, data.remote_dbs, output, mo, iface.remote_db)
            iface.i = 3
        except subprocess.CalledProcessError as ex:
            LOG.error(ex)
        sleep(.05)
        GObject.idle_add(self.prep4)
        return True

    @staticmethod
    def prep2(db_info, output, mo):
        db_info.set_tooltip_text('Underlining indicates a local database')
        for db in output:
            db = db.split('\t')
            try:
                mo.append([db[0], Path(db[0]).name, human_bytes(
                    sum(f.stat().st_size for f in Path(db[0]).parent
                        .glob('**/*') if f.is_file())), db[1], True])
            except IndexError:
                pass  # if directory does not exist or is empty

    @staticmethod
    def prep3(remote_dbs, output, mo, remote_db):
        remote_dbs.clear()
        for i, db in enumerate(output.strip().split('\n')[1:]):
            db = db.split('\t')
            # fill db_info table
            mo.append([None, db[0], human_bytes(
                2 ** 30 * float(db[2].strip())), db[1], False])
            # fill remote_db ComboBox
            remote_dbs.append([db[0], i])
            if db[0] == 'nt':
                remote_db.set_active(i)
        if len(remote_dbs) == 0:
            remote_dbs.append(['no remotes found', 0])

    def prep4(self):
        iface = self.iface
        iface.thread.join()
        # sync db selection in db_info TreeView and blast_db ComboBox
        sel = iface.db_info.get_selection()
        sel.connect('changed', lambda *args: iface.blast_db.
                    set_active(sel.get_selected_rows()[1][0].get_indices()[0]))
        sel.select_path(0)
        iface.blast_db.connect('changed', lambda *args: sel
                               .select_path(iface.blast_db.get_active()))
        LOG.info('BLAST prep done')
        return False

    def refresh(self):
        """Re-view the page. Re-load the current genes to the marker-gene selector"""
        data = self.data
        iface = self.iface

        if not iface.blast_seen:
            # look for BLAST+ executable in the config and on the $PATH; fill db_info table
            for path in [shutil.which('blastn'), data.blast_path]:
                if self.prep0(path):
                    break
            iface.blast_seen = True

        iface.blast_gene.remove_all()
        genes = list(data.genes)
        [iface.blast_gene.append_text(gene) for gene in genes]
        if data.gene_for_preview and data.gene_for_preview in genes:
            idx = genes.index(data.gene_for_preview)
        else:
            idx = 0
        iface.blast_gene.set_active(idx)

        # always allow skipping BLAST
        self.set_changed(PAGE, False)

    def start_BLAST(self, widget, mode, *args):
        """Set-up the BLAST thread"""
        data = self.data
        iface = self.iface

        if 'blast_wrapper' in iface:
            # The button was pressed when it was in the 'Stop' state
            iface.pill2kill.set()
            sleep(.1)
            if iface.blaster.is_alive():
                sleep(.1)
                iface.blaster.stop()
            return

        iface.blast_spinner.start()
        iface.spinbox.set_visible(True)
        gene = iface.blast_gene.get_active_text()
        data.gene_for_preview = gene
        LOG.info('running %s for %s' % (mode, gene))
        if mode == 'local':
            fasta = self.wd / gene / (gene + '.fasta')
            if not fasta.is_file():
                self.show_notification('%s with sequences for BLAST does not exist' % fasta)
                iface.blast_spinner.stop()
                iface.spinbox.set_visible(False)
                return
            pars = [True, False, False]
            adapt_button = iface.blast_local, mode
            # re-read all sequence data
            seqdata = {g: {r.id: r for r in SeqIO.parse(
                self.wd / g / (g + '_all.fasta'), 'fasta')} for g in data.genes}

        elif mode == 'remote':
            fasta = iface.missing_fasta.get_filename()
            if not fasta and mode == 'remote':
                self.show_notification('select a FASTA with sequence data first')
                iface.blast_spinner.stop()
                iface.spinbox.set_visible(False)
                return
            pars = [False, True, False]
            adapt_button = iface.blast_remote, mode
            # read seqdata from selected file
            seqdata = {gene: {r.id: r for r in SeqIO.parse(fasta, 'fasta')}}

        else:  # XML import
            pars = [True, True, iface.xml_import.get_filenames()]
            adapt_button = iface.blast_import, 'Load'
            # re-read all sequence data
            seqdata = {g: {r.id: r for r in SeqIO.parse(
                self.wd / g / (g + '_all.fasta'), 'fasta')} for g in data.genes}

        pars = dict(zip(['no_remote', 'no_local', 'BLAST_xml'], pars))
        (im, la), tx = adapt_button[0].get_child().get_children(), adapt_button[1]
        im.set_from_icon_name('media-playback-stop-symbolic', 4)
        la.set_text('Stop')
        db_row = iface.blast_db.get_model()[iface.blast_db.get_active()]

        # define static parameters
        args = {'gui': True,
                'cfg': ab12phylo_app_base.CFG,
                'no_BLAST': False,
                'genes': [gene],
                'db': db_row[1],
                'remote_db': data.remote_dbs[iface.remote_db.get_active()][0],
                'timeout': repo.DOWNLOAD_TIMEOUT,
                'dir': self.wd,
                'xml': self.wd / repo.PATHS.xml,
                'www_xml': str(self.wd / repo.PATHS.www_xml),
                'tsv': self.wd / repo.PATHS.tsv,
                'bad_seqs': self.wd / repo.PATHS.bad_seqs,
                'missing_fasta': self.wd / repo.PATHS.missing_fasta,
                'dbpath': Path(db_row[0]).parent if db_row[0] else None}
        args.update(pars)
        args = Namespace(**args)
        reader = Namespace(**{'seqdata': seqdata, 'metadata': None})

        iface.blast_tup = (im, la, tx, gene)
        LOG.debug('init blast.blast_build with %s' % str(args))
        iface.blaster = blast.blast_build(args, reader)
        iface.blast_wrapper = threading.Thread(
            # iface.blaster = multiprocessing.Process(
            target=self.do_BLAST)
        GObject.timeout_add(1000, self.update_BLAST)
        iface.blast_wrapper.start()
        return
        # return to main loop

    def do_BLAST(self):
        """Run BLAST thread"""
        with bump_log_level(LOG):
            self.iface.blaster.start()
            self.iface.blaster.join()
        GObject.idle_add(self.stop_BLAST)

    def stop_BLAST(self):
        """Finish the BLAST thread"""
        data = self.data
        iface = self.iface
        if 'blaster' in iface:
            blaster = iface.blaster
            if blaster.update:  # a local database was created
                iface.blast_db.get_model()[iface.blast_db.get_active()][-1] = True
            if iface.blaster.missing_fasta.is_file():
                iface.missing_fasta.set_filename(str(blaster.missing_fasta))
            if blaster.TSV != self.wd / repo.PATHS.tsv:
                shutil.move(blaster.TSV, self.wd / repo.PATHS.tsv)
                # df_old = self._df_with_sp(iface.blaster.TSV)
                # df_old = df_old.loc[df_old['BLAST_species'] != '', :]
                # df_new = self._df_with_sp(self.wd / repo.PATHS.tsv)
                # for _id, series in df_old.iterrows():
                #     df_new.loc[(df_new.index == series.name) & (df_new['gene'] == series.gene), :] = \
                #         df_old.loc[(df_old.index == series.name) & (df_old['gene'] == series.gene), :]

        im, la, tx, gene = iface.blast_tup
        self._refill(gene, fresh=True)

        iface.blast_spinner.stop()
        iface.spinbox.set_visible(False)
        im.set_from_icon_name('media-playback-start-symbolic', 4)
        la.set_text(tx)

        if iface.pill2kill.is_set():
            tx = 'stopped BLAST'
            self.show_notification(msg=tx, secs=2)
        else:
            tx = 'BLAST finished'
            iface.blast_help_stack.set_visible_child_name('sp_info')
            if iface.notebook.get_current_page() != PAGE:
                self.show_notification(msg=tx, secs=5)
            else:
                self.save(silent=True)
        LOG.info(tx)
        iface.pill2kill.clear()
        del iface.blast_wrapper, iface.blaster
        self.set_changed(PAGE, False)

    @staticmethod
    def _save_custom_remote_db(entry, *args):
        combo = entry.get_parent().get_parent()
        if not combo.get_active_iter():
            tx = entry.get_text()
            combo.get_model().append([tx, -1])
            LOG.debug('entered remote_db %s' % tx)

    @staticmethod
    def _df_with_sp(path):
        LOG.debug('reading df')
        df = pd.read_csv(path, sep='\t', dtype={'id': str})
        df = df.set_index('id')
        if 'BLAST_species' in df:
            df.BLAST_species.fillna('', inplace=True)
        if 'extra_species' in df:
            df.extra_species.fillna('', inplace=True)
        if 'pid' not in df:
            df['pid'] = nan
        return df

    def _refill(self, gene=None, fresh=False):
        """Re-fill the species annotation table"""
        data = self.data
        iface = self.iface
        if not gene:
            gene = iface.blast_gene.get_active_text()
        LOG.debug('re-fill species annotations for %s' % gene)
        if 'df' not in iface.tempspace or fresh:
            try:
                iface.tempspace.df = self._df_with_sp(self.wd / repo.PATHS.tsv)
            except FileNotFoundError as ex:
                LOG.debug(ex)
        df = iface.tempspace.df
        data.sp_model.clear()
        df = df.loc[df['gene'] == gene]
        for sample, r in df.iterrows():
            if 'pid' not in r or isnan(r.pid):
                pid, r.BLAST_species, r.extra_species = '', '', ''
            elif type(r.pid) == float:
                pid = '%.2f' % r.pid if r.pid < 100 else '100'
            else:
                pid = r.pid
            color = None
            if 'quality' in r and not pd.isna(r.quality) and r.quality != 'no phreds':
                color = iface.AQUA
            # and again:
            row = [str(i) for i in [sample, pid, r.BLAST_species, r.extra_species]] + [color]
            data.sp_model.append(row)

    def _save_sp_edit(self, cell, path, new_text, tv, col):
        iface = self.iface
        if col == 0:
            return  # do not allow sample IDs to be changed
        if 'df' not in iface.tempspace:
            iface.tempspace.df = self._df_with_sp(self.wd / repo.PATHS.tsv)
        df = iface.tempspace.df
        mo = tv.get_model()
        gene = iface.blast_gene.get_active_text()
        c2 = {1: 'pid', 2: 'BLAST_species', 3: 'extra_species'}
        if col == 1:
            try:
                new_text = float(new_text)
            except ValueError:
                new_text = 0
        # df.loc[df['gene'] == gene].loc[mo[path][0]][c2[col]] = new_text
        # pandas DataFrame get with multiple conditions including index then assign
        df.loc[(df.index == mo[path][0]) & (df['gene'] == gene), c2[col]] = new_text
        self.save_row_edits(cell, path, str(new_text), tv, col)
        self.set_changed(PAGE)


def human_bytes(num):
    step_unit = 1024.0
    for x in ['B', 'KB', 'MB', 'GB', 'TB', 'PB']:
        if num < step_unit:
            return '{:.0f} {}'.format(num, x)
        num /= step_unit


class bump_log_level:

    def __init__(self, log, off=False):
        self.level = max(log.level, logging.INFO)
        self.off = off

    def __enter__(self):
        if not self.off:
            logging.disable(self.level)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)



# 2021 Leo Kaindl

import itertools
import logging
import pickle
import threading
import xml.etree.ElementTree as ET
from copy import deepcopy
from math import ceil
from pathlib import Path
from time import sleep

import gi
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import svgutils.transform as sg
import toyplot
import toytree
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from PIL import Image
from matplotlib.backends.backend_gtk3agg import (
    FigureCanvasGTK3Agg as FigureCanvas)
from matplotlib.cm import get_cmap
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from toyplot import html, pdf, png, svg, color
from toytree.utils import TreeError, ToytreeError

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Gdk, GLib

from ab12phylo import repo
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 7
UI_INFO = """
<ui>
  <popup name='context_menu'>
    <menuitem action='root' />
    <menuitem action='drop' />
    <menuitem action='collapse' />
    <menuitem action='select' />
    <menuitem action='extract' />
    <menuitem action='calculate' />
    <menuitem action='haplo_collapse' />
  </popup>
</ui>
"""


class tree_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface
        phy = data.phy

        # context menu
        action_group = Gtk.ActionGroup(name='context_menu')
        action_group.add_actions(
            [('root', None, 'Root here', None,
              'Root the tree at the selected node(s).', self.tree_modify),
             ('drop', None, 'Drop taxa', None,
              'Drop the selected nodes from the tree.', self.tree_modify),
             ('collapse', None, 'Collapse clade', None,
              'Collapse selected taxa to a leaf.', self.tree_modify),
             ('select', None, 'Select clade', None,
              'Select all descendants of the MRCA.', self.expand_popgen),
             ('extract', None, 'Extract clade', None,
              'Drop all other nodes from the tree.', self.tree_modify),
             ('calculate', None, 'Calculate diversity', None,
              'Calculate diversity metrics for the selected population.', self.start_popgen),
             ('haplo_collapse', None, 'Collapse haplotypes', None,
              'Collapse identical haplotyes in the selected population '
              'to a single leaf.', self.collapse_haplotypes)])

        uimanager = Gtk.UIManager()
        uimanager.add_ui_from_string(UI_INFO)
        self.win.add_accel_group(uimanager.get_accel_group())
        uimanager.insert_action_group(action_group)
        iface.contextmenu = uimanager.get_widget('/context_menu')

        for w_name in ['tree_eventbox', 'view_msa_ids', 'msa_eventbox']:
            iface.__getattribute__(w_name).connect(
                'button-press-event', self._on_right_click, iface.contextmenu)

        iface.query.connect('focus_out_event', self.select_popgen)
        iface.exclude.connect('focus_out_event', self.constrain_selection)

        iface.save_plot.connect('clicked', self.start_save)

        iface.view_msa_ids.set_model(data.tree_anno_model)
        crt = Gtk.CellRendererText(editable=True)
        crt.connect('edited', self._rename_taxon)
        col = Gtk.TreeViewColumn(title='id', markup=0, foreground=2, background=3,
                                 cell_renderer=crt)
        col.props.max_width = 80
        iface.view_msa_ids.append_column(col)
        iface.view_msa_ids.set_tooltip_column(1)

        iface.pop_metrics.set_model(data.pop_model)

        # set some plot_menu MenuButton images
        iface.flip.set_image(iface.flipim)
        iface.dist.set_image(iface.distim)
        iface.anno.set_image(iface.annoim)
        iface.filetypes.set_image(iface.saveim)
        iface.space.set_image(iface.spaceim)

        # connect zooming and selecting
        iface.tree_sel = iface.view_msa_ids.get_selection()
        iface.tree_sel.set_mode(Gtk.SelectionMode.MULTIPLE)
        iface.tree_sel.connect('changed', self.keep_visible, iface.view_msa_ids,
                               iface.parallel_tree.props.vadjustment.props,
                               iface.tempspace)
        iface.msa_eventbox.connect_after(
            'button_press_event', self.select_seqs, PAGE, iface.zoomer,
            iface.view_msa_ids, iface.tempspace)  # in-preview selection
        iface.tree_eventbox.connect_after(
            'button_press_event', self.select_seqs, PAGE, iface.zoomer,
            iface.view_msa_ids, iface.tempspace)  # in-preview selection
        iface.tree_pane.connect(
            'scroll-event', self.xy_scale, PAGE)  # zooming

        iface.tree_pane.set_position(599)
        self.iface.ratio = 600 / self.win.get_default_size()[0]
        self.win.connect_after('window-state-event', self._size_pane)

        iface.pane_handler = iface.tree_pane.connect_after(
            'notify::position', self._size_scrollbars,
            iface.tree_left_scrollbar, iface.tree_right_scrollbar,
            iface.tree_spacer, iface.view_msa_ids)

        iface.subtree.connect('clicked', self.expand_popgen)
        iface.calculate.connect('clicked', self.start_popgen)

        phy.tx = 'circular' if phy.circ else 'rectangular'
        phy.tx += '_TBE' if phy.tbe else '_FBP'
        phy.array = None
        iface.tree = None

    def reset_tree(self, *args):
        """
        Re-set the tree modifications by deleting
        the dictionary and the modified tree
        """
        if self.iface.thread.is_alive():
            self.iface.thread.join()
        self.data.phy.modify = dict()
        if 'g_lens' in self.data.phy:
            del self.data.phy.g_lens
        if (self.wd / repo.PATHS.modified_tree).is_file():
            (self.wd / repo.PATHS.modified_tree).unlink()
        self.start_phy()

    def refresh(self):
        """Re-view the page"""
        data = self.data
        iface = self.iface
        phy = data.phy
        self.reload_ui_state(page=PAGE)
        iface.tree_pane.set_position(600)

        if 'tx' in phy and (self.wd / phy.tx).with_suffix('.png').is_file() \
                and (self.wd / repo.PATHS.phylo_msa).is_file() \
                and 'shape' in phy:
            try:
                # load tree PNG
                self.load_image(iface.zoomer, PAGE, iface.tree_eventbox,
                                self.wd / (phy.tx + '.png'), h=phy.shape[1])
                # load MSA PNG
                self.load_image(iface.zoomer, PAGE, iface.msa_eventbox,
                                self.wd / repo.PATHS.phylo_msa,
                                w=phy.shape[0] * self.get_hadj() * 2, h=phy.shape[1])
                # load colorbars
                self.load_colorbar(iface.palplot1)
                if phy.axis and len(data.genes) > 1:
                    self.load_colorbar(iface.gbar, gbar=True)
                    iface.gbar.set_visible(True)
                self.win.show_all()
                return
            except GLib.Error:
                pass

        self.start_phy()
        self.win.show_all()

    def select_popgen(self, widget, ev):
        """Select samples for popgen calculations by name / species matching"""
        data = self.data
        iface = self.iface
        phy = data.phy
        phy.query = widget.get_text()
        phy.exclude = iface.exclude.get_text()

        if not phy.query and not phy.exclude:
            return
        mo = data.tree_anno_model

        search_space = [_id + sp for _id, sp in mo.get_column((0, 1))]
        sel = list()

        # first match
        if not phy.query:
            sel = list(range(len(mo)))
        else:
            sel = [i for i, label in enumerate(search_space)
                   if repo.rgx(phy.query).match(label)]

        # then exclude
        if phy.exclude:
            sel = [i for i, label in enumerate(search_space) if
                   i in sel and not repo.rgx(phy.exclude).match(label)]

        # then select in the model
        iface.tree_sel.unselect_all()
        [iface.tree_sel.select_path(i) for i in sel]

    def constrain_selection(self, widget, ev, passed=None):
        data = self.data
        iface = self.iface
        phy = data.phy
        if passed:
            sel_idx = passed
        else:
            sel_idx = [tp.get_indices()[0] for tp
                       in iface.tree_sel.get_selected_rows()[1]]
        if not sel_idx:
            return

        # get exclusion search space
        idx_and_label = [(idx, _id + sp) for idx, (_id, sp) in enumerate(
            data.tree_anno_model.get_column((0, 1))) if idx in sel_idx]

        # then exclude
        phy.exclude = iface.exclude.get_text()
        sel_idx = [i for i, label in idx_and_label if
                   not repo.rgx(phy.exclude).match(label)]
        if widget:
            # then select in the model
            iface.tree_sel.unselect_all()
            [iface.tree_sel.select_path(i) for i in sel_idx]
        else:
            return sel_idx

    def expand_popgen(self, widget, sel_idx=None):
        """Select a subtree starting from a GtkTreeSelection"""
        data = self.data
        iface = self.iface
        phy = data.phy
        if not sel_idx:
            sel_idx = [tp.get_indices()[0] for tp
                       in iface.tree_sel.get_selected_rows()[1]]
        if not sel_idx:
            self.show_notification('Cannot expand to subtree, no selection', secs=1)
            return
        if not self.iface.tree:
            self.show_notification('No tree in memory, re-drawing first', secs=5)
            self.start_phy(run_after=(self.expand_popgen, (None, sel_idx)))
            return

        in_blue = lambda c: c.replace('style="normal"', 'style="fill:%s"'
                                      % repo.tohex(repo.rocket[0]))

        # get a list of all tip labels and whether they are selected
        id_and_ifsel = [(in_blue(_id), i in sel_idx) for i, _id in enumerate(
            iface.tree.get_tip_labels()[::-1])]

        # move up in the tree: find the MRCA
        mrca = iface.tree.get_mrca_idx_from_tip_labels(
            names=[_id for _id, sel in id_and_ifsel if sel])

        # move down in the tree: get tip positions
        sel_idx = [[_id for _id, sel in id_and_ifsel].index(_id)
                   for _id in iface.tree.get_tip_labels(mrca)]

        if phy.exclude:
            sel_idx = self.constrain_selection(None, None, sel_idx)

        [iface.tree_sel.select_path(i) for i in sel_idx]
        iface.pop_size.set_text('population size: %d' % len(sel_idx))
        iface.pop_size.set_visible(True)

    def start_popgen(self, widget=None, haplo_collapse=False):
        data = self.data
        iface = self.iface
        phy = data.phy
        data.pop_model.clear()
        self.init_popgen_columns(iface.pop_metrics)
        sleep(.1)  # prevent crash on re-load->Ctrl+A->calculate

        mo, tps = iface.tree_sel.get_selected_rows()
        tps = [tp[0] for tp in tps]
        if haplo_collapse:
            tps = haplo_collapse

        if len(tps) < 2:
            self.show_notification('invalid selection', secs=1)
            return

        if phy.array is None:
            self.start_phy()
            return

        for w_name in ['gap_share', 'unk_share']:
            phy.__setattr__(w_name, iface.__getattribute__(
                w_name).get_adjustment().props.value)

        # re-direct to thread
        iface.i = 0
        iface.k = len(data.genes) + 2
        iface.text = 'calculating'
        iface.thread = threading.Thread(target=self.do_popgen, args=[tps, haplo_collapse])
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()

    def do_popgen(self, tps, haplo_collapse):
        data = self.data
        iface = self.iface
        phy = data.phy
        iface.pop_size.set_text('population size: %d' % len(tps))
        iface.pop_size.set_visible(True)

        # show the ids of the selected samples
        buf = Gtk.TextBuffer()
        buf.props.text = ', '.join(
            [_id for i, _id in enumerate(self.data.tree_anno_model
                                         .get_column(0)) if i in tps])
        iface.pop_ids.set_buffer(buf)

        # crop the array and shift the entries
        ar = np.flipud(phy.array)[tps, :]
        # shift 4|>7 to 5 and 6|7 to 4
        ar = np.array([[5 if x == 4 or x > 7 else (4 if x in {6, 7} else x)
                        for x in _ar] for _ar in ar])
        start, overall, drops, repeats = 0, set(), set(), dict()
        for i, gene in enumerate(data.genes):
            iface.text = 'popgen: %s' % gene
            LOG.debug(iface.text)
            _range = range(start, start + phy.g_lens[gene])
            overall.update(_range)
            row, drops, repeats = _per_gene_diversity(gene, phy, ar, _range)
            data.pop_model.append(row)
            start += phy.g_lens[gene] + len(repo.SEP)
            iface.i += 1

        if len(data.genes) > 1:
            iface.text = 'popgen: overall'
            LOG.debug(iface.text)
            row, drops, repeats = _per_gene_diversity('overall', phy, ar, overall)
            data.pop_model.append(row)
            iface.i += 1

        # show the table
        iface.tree_stack.set_visible_child_name('popgen')

        self._highlight_duplicates(tps, repeats)

        if haplo_collapse:
            self.collapse_haplotypes(None, return_loop=True)
        else:
            # plot using tps and drops
            self._do_phy5((tps, drops, repeats))

    def _highlight_duplicates(self, tps, repeats):
        """
        Highlight duplicate sequences by shading the ID column in matching color.
        :param tps: indices [int] of selected tree tips
        :param repeats: {idx: [occurrences of seq found at idx, given it occurs several times]}
        :return:
        """
        mo = self.data.tree_anno_model
        # remove old color markup
        for row in mo:
            if row[3] != self.iface.AQUA:
                row[3] = None

        colors = [repo.tohex(c[:3]) for c in get_cmap('Set1', len(repeats)).colors]
        for color, idcs in zip(colors, repeats.values()):
            for idx in idcs:
                mo[tps[idx]][3] = color

    def collapse_haplotypes(self, action, return_loop=False, sel_idx=None):
        if not return_loop:
            if not sel_idx:
                sel_idx = [tp.get_indices()[0] for tp
                           in self.iface.tree_sel.get_selected_rows()[1]]
            if not sel_idx:
                self.show_notification('Cannot %s, no selection' % action.get_name(), secs=2)
                return
            if not self.iface.tree:
                self.show_notification('No tree in memory, re-drawing first', secs=5)
                self.start_phy(run_after=(self.collapse_haplotypes, (action, False, sel_idx)))
                return
            # first re-find the haplotyes
            self.start_popgen(haplo_collapse=sel_idx)
        else:
            # then collapse
            props = self.iface.spacespin.get_adjustment().props
            props.value = max(2.4, props.value)
            mod = self.data.phy.modify
            candidates = dict()
            for i, (n, sp, c) in enumerate(self.data.tree_anno_model.get_column((0, 1, 3))):
                if not c or c == self.iface.AQUA:
                    continue
                if sp.startswith('REF_'):
                    n = sp.split(':')[0]
                if c in candidates:
                    candidates[c][i] = n  # tips[i]
                else:
                    candidates[c] = {i: n}  # tips[i]}
            if candidates:
                if 'rename' not in mod:
                    mod['rename'] = dict()
                if 'drop' not in mod:
                    mod['drop'] = list()
            for haplo_dict in candidates.values():
                # rename the remaining node
                mod['rename'][haplo_dict[min(haplo_dict)]] = '~'.join(
                    sorted([t for t in haplo_dict.values() if not t.startswith('REF_')]) + \
                    sorted([t for t in haplo_dict.values() if t.startswith('REF_')]))
                # delete the others
                mod['drop'].extend([v for k, v in haplo_dict.items() if k != min(haplo_dict)])

            self.iface.run_after = self.start_phy, ()

            sleep(.1)
            GObject.idle_add(self._stop_phy)
            return True

    def _rename_taxon(self, cell, path, new_text):
        mo = self.data.tree_anno_model
        old_text = mo[path][0]
        if mo[path][1].startswith('REF_'):
            old_text = mo[path][1].split(':')[0]
        if old_text == new_text:
            return
        elif new_text in mo.get_column(0) + [i.split(':')[0] for i in mo.get_column(1)]:
            return
        LOG.debug('renaming %s to %s' % (old_text, new_text))
        mo[path][0] = new_text
        new_text = new_text.replace(', ', '~').replace(',', '~')
        old_text = old_text.replace(', ', '~').replace(',', '~')
        mod = self.data.phy.modify
        if 'rename' not in mod:
            mod['rename'] = {old_text: new_text}
        else:
            # save twice
            for k, v in mod['rename'].items():
                if v == old_text:
                    mod['rename'][k] = new_text
            mod['rename'][old_text] = new_text
        LOG.debug(mod['rename'])

    def init_popgen_columns(self, tv):
        if tv.get_n_columns() > 0:
            return
        for i, ti in enumerate(['gene', 'sites', 'S', 'k', 'π', 'Watterson\'s θ',
                                'Tajima\'s D', 'unique seqs', 'gaps', 'unknown']):
            col = Gtk.TreeViewColumn(
                title=ti, cell_renderer=Gtk.CellRendererText(), text=i)
            col.set_resizable(True)
            tv.append_column(col)
        tv.realize()
        tv.columns_autosize()

    def init_sp_genes(self, sp_genes, genes, sel_gene=None):
        """
        Init the selector for which gene to draw species annotations from
        :param sp_genes: the GtkComboBoxText to display the genes
        :param genes: the iterable of genes in the analysis
        :param sel_gene: an optional selected gene
        :return: sel_gene: if not set before, is 'best pid' now
        """
        sp_genes.remove_all()
        genes = list(genes)
        if len(genes) > 1:
            genes.insert(0, 'best pid')
        [sp_genes.append_text(gene) for gene in genes]
        if sel_gene and sel_gene in genes:
            idx = genes.index(sel_gene)
        else:
            idx = 0
            sel_gene = 'best pid'
        sp_genes.set_active(idx)
        return sel_gene

    def start_save(self, *args):
        """Handle the "Export the current plot now" save functionality via a thread."""
        iface = self.iface
        phy = self.data.phy
        if 'canvas' not in iface:
            self.start_phy()
            return

        # parse plot settings only when needed
        [phy.__setattr__(w_name, iface.__getattribute__(
            w_name).get_active()) for w_name in
         ['rect', 'circ', 'unro', 'tbe', 'fbp', 'supp', 'spec',
          'axis', 'align', 'pmsa', 'pdf', 'svg', 'png', 'nwk', 'html']]
        phy.flip = iface.flipspin.props.adjustment.props.value
        phy.dist = iface.distspin.props.adjustment.props.value
        phy.space = iface.spacespin.props.adjustment.props.value
        phy.sel_gene = iface.sp_genes.get_active_text()

        # re-direct to thread
        iface.thread = threading.Thread(target=self._do_save)
        iface.k = 2 + 2 * phy.pdf + 2 * phy.svg + phy.pmsa + phy.html
        iface.i = 0
        iface.text = 'read tree'
        sleep(.1)
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        return
        # return to main loop

    def _do_save(self, *args):
        iface = self.iface
        self.save_tree()
        self.save_msa()
        self.iface.text = 'idle'
        sleep(.1)
        GObject.idle_add(self._stop_save)
        return True

    def _stop_save(self, *args):
        self.iface.thread.join()
        LOG.debug('phylo saving idle')
        self.win.show_all()
        self.data.change_indicator[PAGE] = False
        return False

    def save_tree(self, *args):
        """Save the current tree in different formats"""
        data = self.data
        iface = self.iface
        phy = data.phy
        LOG.debug('rendering tree')
        if 'canvas' not in iface:
            self.start_phy()
            return
        tx = self.wd / phy.tx
        iface.text = 'rendering png'
        png.render(iface.canvas, str(tx.with_suffix('.png')), scale=1.6)
        iface.i += 1
        if phy.pdf:
            iface.text = 'rendering pdf'
            pdf.render(iface.canvas, str(tx.with_suffix('.pdf')))
            iface.i += 1
        if phy.svg:
            iface.text = 'rendering svg'
            svg.render(iface.canvas, str(tx.with_suffix('.svg')))
            iface.i += 1
        if phy.html:
            iface.text = 'rendering html'
            html.render(iface.canvas, str(tx.with_suffix('.html')))
            iface.i += 1

    def save_msa(self, popgen=None):
        """Save the current msa in different formats"""
        data = self.data
        iface = self.iface
        phy = data.phy
        LOG.debug('rendering msa')
        Path.mkdir(self.wd / repo.PATHS.phylo_msa.parent, exist_ok=True)
        msa = self.wd / repo.PATHS.phylo_msa.parent / repo.PATHS.phylo_msa.stem
        dpi = iface.scale * repo.DPI
        iface.text = 'rendering png'
        iface.fig.savefig(msa.with_suffix('.png'), transparent=True, dpi=dpi)
        iface.i += 1

        if popgen:
            return
        if phy.svg:
            iface.text = 'rendering svg with msa -- SLOW'
            iface.fig.savefig(msa.with_suffix('.svg'), transparent=True)
            iface.i += 1
        if phy.pdf:
            iface.text = 'rendering pdf with msa -- SLOW'
            iface.fig.savefig(msa.with_suffix('.pdf'), transparent=True, dpi=dpi)
            iface.i += 1

        if phy.pmsa:
            if phy.svg:
                iface.text = 'collating svg tree and msa -- SLOW'
                LOG.info(iface.text)
                tree = sg.fromfile((self.wd / phy.tx).with_suffix('.svg'))
                m = sg.fromfile(msa.with_suffix('.svg'))
                mw, mh = [float(i[:-2]) * 96 / 72 for i in m.get_size()]  # now in px
                pt = tree.getroot()
                tw, th = [float(i[:-2]) for i in tree.get_size()]  # already in px
                pm = m.getroot()
                pm.moveto(tw, 0, th / mh / 6, th / mh * 96 / 72)
                fig = sg.SVGFigure()  # '%.0fpx' % (tw + th / mh / 8), '%.0fpx' % th)
                fig.append([pt, pm])
                fig.save(self.wd / (phy.tx + '_msa.svg'))
            else:
                iface.text = 'rendering png with msa'
                LOG.debug(iface.text)
                old_wi = iface.canvas.width
                wi = phy.array.shape[1] * 3
                im = Image.open(msa.with_suffix('.png'))
                im = im.convert('RGB')
                im = im.resize((wi, int(iface.canvas.height)), resample=Image.NEAREST)
                iface.canvas.width += wi
                im_axes = iface.canvas.image(im, bounds=(old_wi, old_wi + wi - 20,
                                                         0, iface.canvas.height))
                png.render(iface.canvas, str(self.wd / (phy.tx + '_msa.png')), scale=1.6)
            iface.i += 1

    def start_phy(self, modify=None, run_after=None):
        """Get settings and block GUI"""
        data = self.data
        iface = self.iface
        phy = data.phy
        if iface.thread.is_alive():
            self.show_notification('Busy', secs=1)
            return
        LOG.debug('start_plotting')
        if modify:
            k, v = modify.popitem()
            if k == 'root':  # replace old root
                phy.modify[k] = v
            else:
                phy.modify[k] = phy.modify.get(k, list()) + v

        # empty placeholder early to avoid over-high graphics
        for wi in [iface.tree_eventbox, iface.msa_eventbox]:
            ch = wi.get_child()
            if ch:
                ch.destroy()

        # parse plot settings only when needed
        [phy.__setattr__(w_name, iface.__getattribute__(
            w_name).get_active()) for w_name in
         ['rect', 'circ', 'unro', 'tbe', 'fbp', 'supp', 'spec',
          'axis', 'align', 'pmsa', 'pdf', 'svg', 'png', 'nwk', 'html']]
        phy.flip = iface.flipspin.props.adjustment.props.value
        phy.dist = iface.distspin.props.adjustment.props.value
        phy.space = iface.spacespin.props.adjustment.props.value
        phy.sel_gene = iface.sp_genes.get_active_text()

        # re-direct to thread
        iface.thread = threading.Thread(target=self._do_phy1)
        iface.k = (phy.rect + phy.circ + phy.unro) * 2 + 9 \
                  + 2 + phy.html + phy.pmsa + 2 * phy.svg + 2 * phy.pdf  # saving
        iface.i = 0
        iface.text = 'read tree'
        sleep(.1)
        iface.run_after = run_after
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        return
        # return to main loop

    def _do_phy1(self):
        data = self.data
        iface = self.iface
        phy = data.phy
        drops = set()

        # pick a tree file
        if phy.modify and (self.wd / repo.PATHS.modified_tree).is_file():
            LOG.debug('using modified tree')
            tree_file = self.wd / repo.PATHS.modified_tree
        else:
            tree_file = self.wd / repo.PATHS.fbp if phy.fbp \
                else self.wd / repo.PATHS.tbe
            if not tree_file.is_file():
                GObject.idle_add(self._stop_phy, 'no tree found at %s' % str(tree_file))
                sleep(.1)
                return True

        # read tree file
        tr = toytree.tree(open(tree_file, 'r').read(), tree_format=0)

        if 'collapse' in phy.modify:
            names = [i for i in phy.modify['collapse'] if i in tr.get_tip_labels()]
            if names:
                # move references to the back
                names = sorted([t for t in names if not t.startswith('REF_')]) \
                        + sorted([t for t in names if t.startswith('REF_')])
                try:
                    mrca_idx = tr.get_mrca_idx_from_tip_labels(names=names)
                    mrca = tr.idx_dict[mrca_idx]
                    drops = set(mrca.get_leaf_names()) - set(names)
                    [t.detach() for t in mrca.get_children()]
                    mrca.name = '~'.join(names)
                    phy.modify['collapse'] = [' ']  # only remember that, not what
                    tr = toytree.tree(tr.newick)
                except (TreeError, ToytreeError) as te:
                    GObject.idle_add(self._do_phy2, [str(te)])
                    sleep(.1)
                    return True

        if 'drop' in phy.modify:
            names = [i for i in phy.modify['drop'] if i in tr.get_tip_labels()]
            if names:
                tr = tr.drop_tips(names=names)
                tr = toytree.tree(tr.newick)

        if 'root' in phy.modify:
            try:
                names = [i for i in phy.modify['root'] if i in tr.get_tip_labels()]
                if names:
                    tr = tr.root(names=names)
            except ToytreeError as te:
                GObject.idle_add(self._do_phy2, [str(te)])
                sleep(.1)
                return True

        if 'extract' in phy.modify:
            names = [i for i in phy.modify['extract'] if i in tr.get_tip_labels()]
            if names:
                treenode = tr.idx_dict[tr.get_mrca_idx_from_tip_labels(names=names)]
                tr = toytree.tree(treenode.write())

        if 'rename' in phy.modify:
            for old, new in phy.modify['rename'].items():
                try:
                    tr.idx_dict[tr.get_mrca_idx_from_tip_labels(names=[old])].name = new
                except (TreeError, ToytreeError) as te:
                    pass

        if phy.modify:
            # save the modified tree
            tr.write(self.wd / repo.PATHS.modified_tree, tree_format=0)

        # fetch tips labels, now top-to-bottom
        phy.tips = list(reversed(tr.get_tip_labels()))
        separated_tips = {item for t in phy.tips for item in t.split('~')}
        iface.i += 1
        iface.tree = tr

        iface.text = 'read metadata'
        # read in tabular data
        df = pd.read_csv(self.wd / repo.PATHS.tsv, sep='\t', dtype={'id': str})
        df.set_index('id', inplace=True)
        iface.tempspace.df = df
        df = df.copy(deep=True)

        r, b, e = 'reference_species', 'BLAST_species', 'extra_species'
        if r in df:
            # overwrite BLAST with reference
            df.loc[df[r].notna(), b] = df.loc[df[r].notna(), r]
        if b in df:
            df = df.rename(columns={b: 'species'})
            phy.did_BLAST = True
        # crop to relevant columns
        df = df[[c for c in ['gene', 'accession', 'species', 'pid'] if c in df]]

        # carry over renamed taxa to metadata entries
        if 'rename' in phy.modify:
            df = df.reset_index()
            for old_id, new_id in phy.modify['rename'].items():
                dd = df.loc[(df['id'] == old_id)].copy()
                dd.loc[:, 'id'] = new_id
                df = df.append(dd)
            df.set_index('id', inplace=True)

        df = df.fillna(value={'pid': 0})
        pid = 'pid' in df
        if phy.sel_gene not in [None, 'best pid'] or len(data.genes) == 1:
            # picked species labels for a single, unique gene
            df = df.loc[df['gene'] == phy.sel_gene]
            df = df.loc[[t for t in separated_tips if t in df.index]]  # skip dropped nodes
            phy.ndf = df.to_dict(orient='index')
        else:
            # picked species labels from best pid across several genes
            # split into per-gene dataframes, filtering out filtered out genes
            dfs = [i[1] for i in df.groupby(df['gene']) if i[0] in data.genes]
            # find row with highest pid from all genes
            phy.ndf = dict()
            for t in separated_tips:
                if any([t not in d.index for d in dfs]):
                    continue  # a dropped node
                rows = [d.loc[t] for d in dfs]
                if pid:
                    phy.ndf[t] = [dict(r) for r in rows if r.pid ==
                                  max([q.pid for q in rows])][0]
                else:
                    phy.ndf[t] = [dict(r) for r in rows][0]

        # fetch species labels for compound nodes
        for compound_tip in set(phy.tips) - separated_tips:
            df = pd.DataFrame.from_records(
                phy.ndf[t] for t in compound_tip.split('~') if t in phy.ndf)
            if 'accession' in df:
                # extract simple species label
                df.loc[df['accession'].notna(), 'species'] = \
                    df.loc[df['accession'].notna()].species.replace(' strain.+', '', regex=True)
            if pid:
                df = df.loc[df.pid == df.pid.max()]
                # split compound species, then re-collate
                species = ' / '.join({item for sp in set(df.species) for item in sp.split(' / ')})
                phy.ndf[compound_tip] = {'species': species, 'pid': df.pid.max()}
            else:
                phy.ndf[compound_tip] = dict()

        iface.i += 1
        GObject.idle_add(self._do_phy2, None, drops)
        sleep(.1)
        return True

    def _do_phy2(self, errors=None, drops=None):
        """Bounce to idle to size the id column"""
        data = self.data
        iface = self.iface
        phy = data.phy
        iface.thread.join()
        in_blue = '<span style="normal">%s</span>'

        if errors:
            iface.text = 'idle'
            self.show_notification('Tree modification failed', errors)
            return
        if drops:
            self.show_notification('Collapsing a branch removed '
                                   'taxa that were not selected', drops, secs=6)

        iface.text = 'fill id column'
        LOG.debug(iface.text)
        data.tree_anno_model.clear()

        # write samples to model
        for t in phy.tips:
            if t not in phy.ndf and '~' not in t:
                data.tree_anno_model.append([t, '', None, iface.AQUA])
                continue  # a dropped node
            nd = phy.ndf[t]
            # make sure pid field is present
            pid = nd.get('pid', 0)
            if pd.isna(pid):
                pid = 0
            nd['pid'] = pid
            # make sure species field is present
            sp = nd.get('species', '')
            if pd.isna(sp):
                sp = ''
            nd['species'] = sp
            if 'accession' in nd and not pd.isna(nd['accession']):
                sp = t + ': ' + sp
                if len(data.genes) > 1 and 'strain' in nd['species']:
                    t = nd['species'][nd['species'].find('strain') + 7:]
                else:
                    t = nd['accession']
                c = iface.BLUE
            else:
                c = None  # iface.FG

            # edit compound tips
            if '~' in t:
                t = t.split('~')
                # for i, tp in enumerate(t):
                #     if tp.startswith('REF_'):
                #         nd = phy.ndf[tp]
                #         if len(data.genes) > 1 and 'strain' in nd['species']:
                #             t[i] = in_blue % nd['species'].split(' strain ')[1]
                #         else:
                #             t[i] = in_blue % nd['accession']
                t = ', '.join(t)
            data.tree_anno_model.append([t, sp, c, None])

        iface.view_msa_ids.set_margin_bottom(50 * phy.axis)
        iface.view_msa_ids.columns_autosize()
        iface.view_msa_ids.realize()
        sleep(.1)

        # re-direct to thread
        iface.thread = threading.Thread(target=self._do_phy3)
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()

    def _do_phy3(self):
        """Write new, annotated MSA, annotated nwk tree and plot tree"""
        data = self.data
        iface = self.iface
        phy = data.phy
        iface.text = 'write new MSA'
        errors = list()

        # read MSA
        records = {r.id: r.seq for r in SeqIO.parse(self.wd / repo.PATHS.msa, 'fasta')}
        # carry over renamed taxa to seqdata
        if 'rename' in phy.modify:
            for old_id, new_id in phy.modify['rename'].items():
                if old_id in records:
                    records[new_id] = records[old_id]
        phy.seqdata = list()
        # write re-named re-ordered MSA
        with open(self.wd / repo.PATHS.msa_anno, 'w') as new_msa:
            for tip in phy.tips:
                if tip not in records and '~' not in tip:
                    phy.seqdata.append(None)
                    continue  # a dropped node
                elif '~' in tip:
                    try:
                        ar = np.array([repo.seqtoint(str(records[t]))
                                       for t in tip.split('~')])
                    except KeyError as ke:
                        LOG.exception(ke)
                        GObject.idle_add(self.reset_tree)
                        sleep(.1)
                        return True
                    seq = repo.inttoseq([np.argmax(np.bincount(ar[:, i])) for i in range(ar.shape[1])])
                    records[tip] = seq

                if repo.SEP in records[tip]:
                    # get seq and cut out artificial separator
                    seq = str(records[tip]).replace(repo.SEP, '')
                elif 'g_lens' in phy:
                    # introduce separators for the sake of plotting
                    seq = str(records[tip])
                    pos = 0
                    for i in range(len(data.genes) - 1):
                        nlen = phy.g_lens[data.genes[i]]
                        seq = seq[:pos + nlen] + repo.SEP + seq[pos + nlen:]
                        pos += nlen + len(repo.SEP)
                else:
                    seq = str(records[tip])

                # get metadata
                try:
                    entry = phy.ndf[tip]
                except KeyError as ke:
                    LOG.exception(ke)
                    GObject.idle_add(self.reset_tree)
                    sleep(.1)
                    return True

                des = entry.get('species', '')
                pid = ('%.2f' % entry['pid']).rstrip('0').rstrip('.')
                if des and pid != '0':
                    des += '; pid %s' % pid

                phy.seqdata.append(repo.seqtoint(records[tip]))

                # rename REFs
                if tip.startswith('REF_') and '~' not in tip:
                    if len(data.genes) > 1 and 'strain' in entry['species']:
                        # # <Strain> <Species> strain <Strain>; pid <pid>
                        # tip = entry['species'].split(' strain ')[1]
                        # <Species> strain <Strain>; pid <pid>
                        tip = des
                        des = ''
                    else:
                        tip = entry['accession']

                SeqIO.write(SeqRecord(Seq(seq), id=tip, description=des), new_msa, 'fasta')
        LOG.debug('wrote updated MSA')
        iface.i += 1

        # get lengths of genes
        if 'g_lens' not in phy:
            phy.g_lens = {gene: len(seq) for gene, seq in
                          zip(data.genes, records[next(iter(records))].split(repo.SEP))}
        multi = True if len(data.genes) > 1 else False

        # cat: REF_ -> 1, no_BLAST_hit -> -1, regular -> 0
        # pid: discrete index in rocket palette
        # confident: 1 if >= flip threshold -> blue, else 0 -> red

        iface.text = 'iterate tree: 1'
        cat_ignore = -1
        in_blue = '<span style="fill:%s">%s</span>' % (repo.tohex(repo.rocket[0]), '%s')

        # guess if support val.s are in [0-1] or in [0-100], then scale with 1 or 1/100 resp.
        sup_scaler = 1 if 1.1 > max([float(s) for s in
                                    iface.tree.get_node_values('support') if s]) else 100
        # rename reference nodes and add features
        for node in iface.tree.treenode.traverse():
            if not node.support:
                node.add_feature('support', 0)
            elif type(node.support) != float:
                node.support = float(node.support)
            if node.is_leaf() and node.name in phy.ndf:
                entry = phy.ndf[node.name]

                if node.name.startswith('REF_'):
                    if multi and 'strain' in entry['species']:
                        entry['species'], strain = entry['species'].split(' strain ')
                        phy.ndf['strain'] = phy.ndf[node.name]
                        node.name = strain
                    else:
                        if 'accession' in entry:
                            acc = entry['accession']
                            phy.ndf[acc] = phy.ndf[node.name]
                            node.name = acc
                        else:  # collapsed node of refs
                            names = node.name.split('~')
                            for i, t in enumerate(names):
                                if t.startswith('REF_'):
                                    ndf = phy.ndf[t]
                                    if multi and 'strain' in ndf['species']:
                                        names[i] = in_blue % ndf['species'].split(' strain ')[1]
                                    else:
                                        names[i] = phy.ndf[t]['accession']
                            node.name = ', '.join(names)
                    node.add_feature('cat', 1)  # blue
                    node.add_feature('species', entry['species'])
                    node.add_feature('pid', 0)  # this is the index of blue in
                    # the rocket palette, which is used to display BLAST pid

                elif 'species' in entry:
                    if '~' in node.name:
                        names = node.name.split('~')
                        for i, t in enumerate(names):
                            if t.startswith('REF_'):
                                ndf = phy.ndf[t]
                                if multi and 'strain' in ndf['species']:
                                    names[i] = in_blue % ndf['species'].split(' strain ')[1]
                                else:
                                    names[i] = in_blue % ndf['accession']
                        node.name = ', '.join(names)
                    node.add_feature('species', entry['species'])
                    node.add_feature('cat', 0 if entry['species'].strip()
                                                 or not phy.spec else 2)  # black else dark_red
                    node.add_feature('pid', 1 + ceil(min(100 - entry['pid'], 20) / 2))  # map to the palette

                else:
                    raise ValueError('No species in ndf dataframe should be impossible')
                    # node.add_feature('cat', 0 if not phy.did_BLAST else 2)  # black else dark_red
                    # node.add_feature('pid', 1 if not phy.did_BLAST else 11)
                    # dark gray if no BLAST, else bad light color signifying no hit

                # meaningless for tips, but to make the fields exist
                node.add_feature('size', 0)
                node.add_feature('confident', 0)
            else:
                # use support values for node size
                node.support /= sup_scaler
                node.add_feature('species', '')
                node.add_feature('confident', 1 if node.support >= phy.flip else 0)
                node.add_feature('pid', -1)  # dark_red for no BLAST hit
                node.add_feature('cat', cat_ignore)  # black, just so the field exists

        iface.i += 2

        if phy.nwk:
            iface.text = 'export nwk'
            exp_tree = deepcopy(iface.tree)
            for node in exp_tree.treenode.traverse():
                if node.is_leaf():
                    try:
                        if node.name in node.species:
                            node.name = node.species
                        else:
                            node.name += ' ' + node.species
                            node.name += ' %.2f' % node.pid
                    except AttributeError:
                        continue
            # write new-ish newick file
            tf = self.wd / repo.PATHS.fbpn if phy.fbp else self.wd / repo.PATHS.tben
            exp_tree.write(tf, tree_format=0)

        # tree plotting tweaks
        iface.text = 'iterate tree: 2'
        for node in iface.tree.treenode.traverse():
            node.dist = max(node.dist, phy.dist)
            if phy.supp:
                node.support = int(round(node.support * 100))
        iface.i += 1

        iface.text = 'prepare plot attributes'
        iface.view_msa_ids.set_margin_bottom(50 * phy.axis)
        iface.view_msa_ids.realize()
        sleep(.1)
        # get model height
        phy.height = iface.view_msa_ids.get_allocated_height() + 50 * phy.axis

        # pre-calculate plot attributes
        # internal node size -> support
        min_size = .2
        ns = [min_size + s / 7.5 for s in iface.tree.get_node_values('support', 1, 1)]

        # internal node color -> support
        red_blue = [color.rgb(*c) for c in [repo.red, repo.blue]]
        nc = [red_blue[i] for i in iface.tree.get_node_values('confident', 1, 1)]

        # primary tip label color -> category
        cat_rgb = [color.rgb(*c) for c in [repo.black, repo.blue, repo.dark_red]]
        tlc = [cat_rgb[n] for n in list(iface.tree.get_node_values(
            'cat', 1, 1)) if n != cat_ignore][::-1]

        # secondary tip label color -> BLAST pid on species
        tl = iface.tree.get_tip_labels()
        if phy.spec:
            # rocket_rgb = [color.rgb(*c) for c in rocket]
            rocket_css = [repo.tohex(c) for c in repo.rocket]
            tl = ['%s <i><span style="fill:%s">%s</span></i>'
                  % (t.name, rocket_css[t.pid], t.species) for t in
                  iface.tree.treenode.traverse('postorder') if t.is_leaf()][::-1]

        # plot tree and msa
        if phy.circ:
            phy.tx = 'circular'
            ly = 'c'
            w, h = 1600, 1600
            iface.canvas = toyplot.Canvas(width=w, height=h)
            iface.axes = iface.canvas.cartesian(bounds=(200, w - 200, 200, h - 200))
        elif phy.rect:
            phy.tx = 'rectangular'
            ly = 'r'
            # get dim for canvas
            w, h = 1200 * phy.space, phy.height  # len(phy.tips) * 14 + 80
            iface.canvas = toyplot.Canvas(width=w, height=h)
            iface.axes = iface.canvas.cartesian(bounds=(20, w, 0, h - 50 * phy.axis),
                                                ymin=-.5, ymax=iface.tree.ntips - .5)
        elif phy.unro:
            raise NotImplementedError
        else:
            raise RuntimeWarning

        phy.tx += '_TBE' if phy.tbe else '_FBP'

        iface.text = 'plotting %s tree' % phy.tx
        with open(self.wd / 'tree.pickle', 'wb') as out_file:
            pickle.dump(iface.tree, out_file)
        try:
            iface.tup = iface.tree.draw(
                axes=iface.axes, scalebar=phy.axis,
                tip_labels=tl, tip_labels_align=phy.align,
                tip_labels_style={'font-size': '15px'},

                node_style={'stroke': None},
                node_sizes=ns if phy.supp else None,
                node_colors=nc,
                node_labels='support' if phy.supp else None,
                node_labels_style={'font-size': '7px', 'fill': '#FFFFFF',
                                   'baseline-shift': '-1px',
                                   'font-weight': 'bold'},

                tip_labels_colors=tlc,
                layout=ly,

                node_hover=phy.html)

            iface.canvas.style['background-color'] = 'white'

            if phy.circ:
                iface.tup[1].show = False
            elif phy.rect:
                iface.tup[1].y.show = False
                iface.tup[1].x.show = phy.axis
                # iface.tup[1].x.spine.position = 'high' # and swap out the 0,h-50 bounds
                th = iface.tree.treenode.height
                iface.tup[1].x.domain.max = (w - 1200) / w  # max(th / 6, phy.spec * th / 2, phy.align * th / 1.6)
                # 1/5 is not enough for aligned labels
                iface.tup[1].x.domain.min = -th
                # 0 is the right-most tip (not tip label) of the tree
                # therefore this inverted x domain.
            iface.i += 1

            try:
                self.save_tree()
            except ValueError:
                LOG.error('canvas to narrow, widening ...')
                iface.spacespin.props.adjustment.props.value += .1

        except ET.ParseError as ex:
            e = 'XML ParseError. Invalid character in a sample ID? Please check metadata.tsv'
            LOG.error(e)
            GObject.idle_add(self.reset_tree)
            sleep(.1)
            return True
        except (ValueError, IndexError) as ex:
            LOG.exception(ex)
            GObject.idle_add(self.reset_tree)
            sleep(.1)
            return True

        GObject.idle_add(self._do_phy4, errors)
        sleep(.1)
        return True

    def _do_phy4(self, errors):
        """Bounce the plotting to the main loop"""
        data = self.data
        iface = self.iface
        phy = data.phy
        iface.thread.join()
        iface.text = 'placing tree'
        LOG.debug(iface.text)
        self.update(PAGE)
        self.win.show_all()
        if errors:
            self.show_notification('Errors during tree plotting', errors)
            return
        sleep(.1)

        self.load_image(iface.zoomer, PAGE, iface.tree_eventbox,
                        self.wd / (phy.tx + '.png'), h=phy.height)
        self.win.show_all()
        sleep(.1)
        iface.i += 1

        # re-direct to thread
        iface.text = 'plotting MSA'
        iface.thread = threading.Thread(target=self._do_phy5)
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()

    def _do_phy5(self, popgen_markup=None):
        """Plot the MSA matrix"""
        data = self.data
        iface = self.iface
        phy = data.phy

        # the tips are saved top-down, but in the tree they're bottom-up. so reverse
        full_len = sum(phy.g_lens.values()) \
                   + (len(data.genes) - 1) * len(repo.SEP)
        for i, seq in enumerate(phy.seqdata):
            if not seq:
                phy.seqdata[i] = [repo.toint('?')] * full_len

        phy.array = np.array(phy.seqdata[::-1])
        # make gaps transparent
        array = np.ma.masked_where(phy.array > repo.toint('else'), phy.array)

        if popgen_markup:
            # make selection non-transparent
            array = np.flipud(array)
            rows, drop_cols, repeats = popgen_markup
            tree_mask = np.full(array.shape, 1.0)
            # mask the unselected rows
            tree_mask[[i for i in range(array.shape[0])
                       if i not in rows], :] = repo.ALPHA
            # mask the invalid columns
            tree_mask[:, sorted(drop_cols)] = repo.ALPHA

        # adjust maximum size
        scale = 6
        while max(array.shape) * scale > 2 ** 14:
            scale -= 1
        LOG.debug('scaling qal with %d' % scale)

        f = Figure()
        f.set_facecolor('none')
        f.set_figheight(array.shape[0] / repo.DPI * 5)
        f.set_figwidth(array.shape[1] / repo.DPI)

        # this line is now redundant
        f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)

        # leave 50px at the bottom for the tree axis and gene legend bar
        b = (50 * phy.axis) / phy.height

        ax = f.add_axes([0, b, 1, 1 - b])
        if popgen_markup:
            mat = ax.matshow(array, alpha=tree_mask, aspect='auto',
                             cmap=ListedColormap(repo.colors),
                             vmin=-.5, vmax=len(repo.colors) - .5)
        else:
            mat = ax.pcolormesh(array, alpha=1, cmap=ListedColormap(repo.colors),
                                vmin=-.5, vmax=len(repo.colors) - .5)
        ax.axis('off')

        # plot a gene marker bar
        if phy.axis and len(data.genes) > 1:
            LOG.debug('adding gene marker bar')
            # build matrix of gene indicators
            gm = [[data.genes.index(g)] * phy.g_lens[g] for g in data.genes]
            # add the spacer
            for i in range(len(data.genes) - 1):
                gm[i] += [-1] * len(repo.SEP)
            gm = [gi for gl in gm for gi in gl]
            # turn into np array
            gm = np.vstack([np.array(gm)] * 2)
            # make spacer transparent
            gm = np.ma.masked_where(gm < 0, gm)
            gene_colors = get_cmap('GnBu', len(data.genes))

            # plot the marker bar onto the MSA graphic
            bax = f.add_axes([0, b * .4, 1, b / 3])
            bar = bax.pcolormesh(gm, cmap=gene_colors)
            # bax.axis('off')
            [bax.spines[t].set_visible(False) for t in ['left', 'right', 'top', 'bottom']]
            bax.yaxis.set_visible(False)
            bax.xaxis.set_ticks_position('bottom')
            bax.tick_params(colors=iface.FG, pad=.2, length=0, labelsize=1)
            bax.xaxis.set_ticks([i for i in range(1, gm.shape[1] - 1) if not i % 100])

            # plot a legend for the gene marker bar
            with plt.rc_context({'axes.edgecolor': iface.FG, 'xtick.color': iface.FG}):
                iface.text = 'gene marker bar'
                LOG.debug(iface.text)
                Path.mkdir(self.wd / repo.PATHS.phylo_msa.parent, exist_ok=True)
                fig = plt.figure(figsize=(4, .2))
                cax = fig.add_subplot(111)
                cbar = ColorbarBase(
                    ax=cax, cmap=gene_colors, orientation='horizontal',
                    ticks=[(.5 / len(data.genes) + j * 1 / len(data.genes))
                           for j in range(len(data.genes))])
                cbar.ax.set_xticklabels(data.genes)
                fig.savefig(self.wd / repo.PATHS.gbar, transparent=True,
                            bbox_inches='tight', pad_inches=0, dpi=600)
                plt.close(fig)
                del fig, cbar

        iface.i += 1
        iface.fig = f
        iface.scale = scale
        self.save_msa(popgen_markup)

        phy.shape = array.shape[1], phy.height

        if iface.rasterize.props.active:
            iface.text = 'place PNG MSA: %d:%d' % array.shape
            LOG.debug(iface.text)
            sleep(.05)
            self.load_image(iface.zoomer, PAGE, iface.msa_eventbox,
                            self.wd / repo.PATHS.phylo_msa,
                            w=phy.shape[0] * self.get_hadj() * 2, h=phy.shape[1])
        else:
            iface.text = 'place vector'
            LOG.debug(iface.text)
            canvas = FigureCanvas(f)  # a Gtk.DrawingArea
            canvas.set_size_request(max(phy.shape[0] * self.get_hadj() * 2, -1),
                                    phy.shape[1])  # height
            try:
                ch = iface.msa_eventbox.get_child()
                if ch:
                    iface.msa_eventbox.remove(ch)
                iface.msa_eventbox.add(canvas)
            except Exception as ex:
                LOG.error(ex)
        plt.close(f)
        del f
        iface.i += 1
        self.load_colorbar(iface.palplot1)

        if phy.axis and len(data.genes) > 1:
            self.load_colorbar(iface.gbar, gbar=True)
            iface.gbar.set_visible(True)
        else:
            iface.gbar.set_visible(False)

        iface.text = 'idle'
        sleep(.1)
        GObject.idle_add(self._stop_phy)
        return True

    def _stop_phy(self, error=None):
        self.iface.thread.join()
        if error:
            LOG.error(error)
            self.show_notification(error, secs=10)
            self.win.show_all()
            return False
        LOG.info('plotting thread idle')
        self.iface.view_msa_ids.set_model(self.data.tree_anno_model)
        self.win.show_all()
        self.save(silent=True)
        self.data.change_indicator[PAGE] = False
        # *Now* go back to where you came from
        if self.iface.run_after:
            func, args, *_ = *self.iface.run_after, (), ()
            func(*args)
            self.iface.run_after = list()
        return False

    def _on_right_click(self, widget, event, menu):
        # Check if right mouse button was pressed
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            menu.popup(None, None, None, None, event.button, event.time)
            return True  # event has been handled

    def _size_scrollbars(self, pane, pos, scl, scr, sp, tv):
        """
        Set a size request for the scrollbar below the narrower
        side of the GtkPaned and set the other one to hexpand.
        """
        p = pane.get_position()
        w = pane.get_allocated_width()
        p = min(p, w)
        self.iface.ratio = p / w
        if p > w / 2:
            scl.set_size_request(-1, -1)
            scl.set_hexpand(True)
            scr.set_size_request(max(w - p - tv.get_allocated_width(), -1), -1)
            scr.set_hexpand(False)
        elif p < w / 2:
            scl.set_size_request(p, -1)
            scl.set_hexpand(False)
            scr.set_size_request(-1, -1)
            scr.set_hexpand(True)
        sp.set_size_request(tv.get_allocated_width(), -1)

    def _size_pane(self, win, state):
        iface = self.iface
        with iface.tree_pane.handler_block(iface.pane_handler):
            try:
                sleep(.1)
                if win.props.is_maximized:
                    s = win.get_screen()
                    m = s.get_monitor_at_window(s.get_active_window())
                    monitor = s.get_monitor_geometry(m)
                    self.iface.tree_pane.set_position(self.iface.ratio * monitor.width)
                else:
                    self.iface.tree_pane.set_position(self.iface.ratio * win.get_size()[0])
            except Exception as ex:
                LOG.error(ex)

    def reload_ui_state(self, *args):
        data = self.data
        iface = self.iface
        phy = data.phy

        for w_name in ['gap_share', 'unk_share']:
            iface.__getattribute__(w_name).get_adjustment().props.value = \
                phy.__getattribute__(w_name)

        for w_name in ['query', 'exclude']:
            iface.__getattribute__(w_name).set_text(
                phy.__getattribute__(w_name))

        # load plot settings
        [iface.__getattribute__(w_name).set_active(
            phy.__getattribute__(w_name)) for w_name in
            ['rect', 'circ', 'unro', 'tbe', 'fbp', 'supp', 'spec',
             'axis', 'align', 'pmsa', 'pdf', 'svg', 'png', 'nwk', 'html']]

        iface.flipspin.props.adjustment.props.value = phy.flip
        iface.distspin.props.adjustment.props.value = phy.dist
        iface.spacespin.props.adjustment.props.value = phy.space
        if data.genes:
            self.init_sp_genes(iface.sp_genes, data.genes, phy.sel_gene)


def _h(a):
    # harmonic number
    b = 0
    for c in range(1, a):
        b += 1 / c
    return b


def _qh(a):
    # quadratic harmonic number
    b = 0
    for c in range(1, a):
        b += 1 / np.power(c, 2)
    return b


def _per_gene_diversity(gene, phy, array, _range):
    conserved, biallelic, polyallelic, unknown, gaps = 0, 0, 0, 0, 0
    start, end = min(_range), max(_range)
    nrows = array.shape[0]
    too_unknown = phy.unk_share * nrows
    too_gappy = phy.gap_share * nrows
    poly = True
    drop = set()

    # iterate over MSA columns, search for segregating sites
    for j in _range:
        if j >= array.shape[1]:
            break
        col = array[:, j]
        if len(set(col)) == 1:
            # only one char at site
            it = col[0]
            if it > 4:
                # all-unknown-site. can happen as we selected a subset of sequences
                unknown += 1
                drop.add(j)
            elif it == 4:
                # all-gap-site
                gaps += 1
                drop.add(j)
            else:
                conserved += 1
        else:  # more than one character in the whole column
            if max(col) > 4 and len(col[col > 4]) > too_unknown:
                # there are too many unknown chars in the column
                unknown += 1
                drop.add(j)
            elif max(col) == 4 and len(col[col >= 4]) > too_gappy:
                # there are too many gaps at the site
                gaps += 1
                drop.add(j)
            else:
                # if gaps or unknown characters are at the site,
                # replace them with the most common nucleotide
                try:
                    col[col >= 4] = np.bincount(col[col < 4]).argmax()
                except ValueError:
                    # there are no nucleotides
                    drop.add(j)
                    continue

                # ordinary site treatment
                num_diff_chars = len(set(col))
                if num_diff_chars > 2:
                    polyallelic += 1
                    if not poly:
                        drop.add(j)
                elif num_diff_chars == 2:
                    biallelic += 1
                else:
                    conserved += 1

    seg_sites = biallelic
    n_sites = end - start + 1 - unknown - gaps
    if poly:
        seg_sites += polyallelic
    else:
        n_sites -= polyallelic

    # crop to allowed sites
    crop = array[:, [i for i in _range if i not in drop]]

    k = 0
    # count pairwise differences
    for combi in itertools.combinations(range(nrows), 2):
        k += np.sum(crop[combi[0]] != crop[combi[1]])

    # binomial coefficient, k = k^ in formula (10) from Tajima1989
    k = k * 2 / nrows / (nrows - 1)
    pi = k / n_sites

    # look up or compute
    if nrows in phy.h:
        a1 = phy.h[nrows]
        a2 = phy.qh[nrows]
    else:
        a1 = _h(nrows)
        phy.h[nrows] = a1
        a2 = _qh(nrows)
        phy.qh[nrows] = a2

    if n_sites == 0:
        return ([gene, n_sites, seg_sites, k, pi, pi, pi,
                 -1, gaps, unknown], drop)

    theta_w = seg_sites / a1 / n_sites  # per site from Yang2014
    unq, idx, count = np.unique(crop, axis=0, return_counts=True, return_index=True)
    repeats = {ri: np.argwhere(np.all(crop == rg, axis=1)).ravel().tolist()
               for rg, ri in zip(unq[count > 1], idx[count > 1])}

    # Tajima's D Formula:
    # d = k - seg_sites/a1 = k - theta_W
    # D = d / sqrt(Var(d))
    try:
        e1 = 1 / a1 * ((nrows + 1) / (3 * nrows - 3) - 1 / a1)
        b2 = (2 * (nrows * nrows + nrows + 3)) / (9 * nrows * (nrows - 1))
        c2 = b2 - ((nrows + 2) / (nrows * a1)) + (a2 / (a1 * a1))
        e2 = c2 / (a1 * a1 + a2)
        tajima = (k - theta_w) / np.sqrt(e1 * seg_sites + e2 * seg_sites * (seg_sites - 1))
    except (ZeroDivisionError, RuntimeWarning) as z:
        tajima = float('inf')

    return ([gene, n_sites, seg_sites, k, pi, theta_w, tajima,
             len(unq), gaps, unknown], drop, repeats)



# 2021 Leo Kaindl

"""
Runs a BLAST+ search on a local database and a remote NCBI BLAST search in the nucleotides
database for seqs missing from it.
Result :code:`XML` files are parsed, and dictionary of non-sequence data is updated and
written to file. Processes are run in a separate thread due to the Python GIL.
"""

import logging
import multiprocessing
import os
import shutil
import sys
import urllib.error
from math import ceil
from os import path
from pathlib import Path
from subprocess import run, check_output, PIPE, CalledProcessError
from time import time, sleep

import pandas as pd
from Bio import SearchIO
from Bio.Blast import NCBIWWW
from numpy import nan

from ab12phylo_cmd import cli, phylo
from ab12phylo_cmd.filter import chmod_x


def _add_xml(*args):
    """Entry point for adding BLAST results"""
    logger = logging.getLogger(__name__)
    # if no argument -> default to cwd
    args = sys.argv[1:]
    if len(args) == 0:
        args = [os.getcwd()]
    args = ['--add_xml'] + args
    # parse to get defaults
    namespace = cli.parser(args, visualize=True).args

    # find and read log
    try:
        log = open(namespace.log[:-5] + '1.log', 'r').read()
    except FileNotFoundError:
        try:
            log = open(namespace.log[:-7] + '.log', 'r').read()
        except FileNotFoundError:
            logger.error('no log file found')
            print('no log file found', file=sys.stderr)
            exit(1)

    # extract genes
    start = log.find('--GENES--') + 10
    end = log[start:start + 800].find('\n')
    namespace.genes = log[start:start + end].split('::')
    logger.debug('read genes from log: %s' % ':'.join(namespace.genes))
    print('read genes from log: %s' % ':'.join(namespace.genes))

    blaster = blast_build(namespace, None)
    print('start reading ...')
    blaster.start()
    # reading an XML requires no waiting
    blaster.join()
    # visualize MSA for inspection
    print('write HTML of MSA ...')
    phylo.mview_msa(namespace)
    print('BYE!')


class blast_build(multiprocessing.Process):
    """
    -none: do nothing
    -xml: read in result files and nothing else
    -local: do not run remote BLAST
    -remote: do not run local BLAST
    If a local search is run:
    Looks for pre-installed BLAST+ 2.9.0 or newer, updates or downloads a database if not
    specified via path, else checks integrity. Runs online Bio.BLAST.NCBIWWW.qblast unless
    prohibited, otherwise checks for earlier result.
    Writes information from best hit to metadata.tsv.
    """

    def __init__(self, _args, reader):
        multiprocessing.Process.__init__(self)
        self._stop_event = multiprocessing.Event()
        self.log = logging.getLogger(__name__)

        # modes / variables
        self.no_BLAST = _args.no_BLAST
        if self.no_BLAST is True:
            return
        self.no_remote = _args.no_remote
        self.no_local = _args.no_local
        self.gene = _args.genes[0]
        self.db = _args.db
        self.remote_db = _args.remote_db
        self.timeout = _args.timeout
        self.gui = 'gui' in _args
        self.df = pd.DataFrame()
        self.update = False

        # Files / Paths
        self.FASTA = path.join(_args.dir, self.gene, self.gene + '.fasta')
        self.XML = _args.xml
        self.www_XML = _args.www_xml
        self.read_XML = _args.BLAST_xml
        self.TSV = _args.tsv
        self.bad_seqs = _args.bad_seqs
        self.missing_fasta = _args.missing_fasta
        self.dbpath = _args.dbpath
        self.blast_dir = None
        os.makedirs(path.join(_args.dir, 'BLAST'), exist_ok=True)
        self.cfg = _args.cfg

        # data in memory
        try:
            self.seqdata = reader.seqdata
            self.metadata = reader.metadata
        except AttributeError as ex:
            self.log.info('passed empty reader')

    def run(self):
        if self.no_BLAST is True:
            self.log.info('--SKIPPING BLAST--')  # leave it alone!
            return
        while not self._stop_event.wait(.1):
            if self.read_XML:
                self.log.info('reading BLAST xml')
                try:
                    not_found = self._parse_remote_result(self.read_XML)
                    self._write_missing(not_found)
                except Exception as ex:
                    self.log.exception(ex)
                return

            if not self.no_local:
                # run local BLAST
                not_found = self._blast_local()
                # if local BLAST failed, continue without annotation
                if not_found is None:
                    return
            else:
                self.log.debug('skip local BLAST')
                # get all IDs for a remote BLAST
                not_found = list(self.seqdata[self.gene].keys())

            self._write_missing(not_found)

            if self.no_remote is True:
                self.log.info('skip online BLAST for missing seqs')
                # try reading old result
                if os.path.isfile(self.www_XML):
                    try:
                        self.log.warning('reading XML result of previous online BLAST')
                        self._parse_remote_result(self.www_XML)
                    except Exception as ex:
                        self.log.error(ex)
                return

            # online BLAST for seqs missing from local db
            self._blast_remote(not_found)
            return

    def stop(self):
        self.terminate()
        self._stop_event.set()
        self.log.warning('killed BLAST')

    def _blast_local(self):
        """
        Optionally check BLAST+ installation, update a local BLAST+ database, check integrity.
        Run a BLAST+ search and finally write updated non-seq data to a .csv file.
        :return: missing sequences as list of sample IDs
        """
        while not self._stop_event.wait(.1):
            # check BLAST+
            try:
                try:
                    binary = self.cfg.get('blastn', shutil.which('blastn'))
                except KeyError:
                    binary = shutil.which('blastn')
                if binary is None:
                    raise ValueError('BLAST+ not installed (not on the $PATH)')
                chmod_x(binary)
                output = check_output(f'{binary} -version', shell=True).decode('utf-8')
                version = [int(i) for i in output.split(',')[0].split(' ')[-1].split('.')]
                if version[0] == 2 and version[1] >= 9 or version[0] > 2:
                    self.blast_dir = path.dirname(binary)
                    blast_dir = Path(self.blast_dir)
                else:
                    raise ValueError(f'BLAST+ in {self.blast_dir} is outdated')
            except (CalledProcessError, ValueError) as e:
                self.log.error(e)
                exit(e)

            # if BLAST db path was provided, use and don't update
            if self.dbpath is not None:
                self.update = False
            else:
                # save BLAST db in root directory
                self.dbpath = path.join(path.dirname(path.dirname(__file__)), 'blastdb', self.db)
                self.update = True
                os.makedirs(self.dbpath, exist_ok=True)

            # update database
            if self.update:
                self.log.debug('downloading or updating BLAST+ db ...')
                binary = blast_dir / 'update_blastdb.pl'
                chmod_x(binary)
                arg = f'{binary} --decompress {self.db} --passive --timeout {self.timeout}'
                try:
                    p = run(arg, shell=True, stdout=PIPE, cwd=self.dbpath)
                    self.log.debug(p.stdout.decode('utf-8').strip())
                except CalledProcessError as e:
                    self.log.exception('BLAST+ db update failed, returned code %d' % e.returncode)

            # check database
            self.log.debug('checking %sBLAST db %s in %s'
                           % ('user-supplied ' if not self.update else '', self.db, self.dbpath))
            binary = blast_dir / 'blastdbcmd'
            chmod_x(binary)
            db = path.join(self.dbpath, self.db)
            arg = f'{binary} -db "{db}" -info'
            try:
                res = check_output(arg, shell=True).decode('utf-8')
                if 'error' in res:
                    raise ValueError
            except (CalledProcessError, ValueError) as e:
                self.log.error('Error in BLAST+ db')
                os._exit(1)  # this is no ordinary exit; it kills zombies, too!
            self.log.debug('BLAST+ db was ok')

            # run BLAST+
            self.log.debug('BLASTing locally ...')
            binary = blast_dir / 'blastn'
            chmod_x(binary)
            arg = f'{binary} -db "{db}" -query "{self.FASTA}" -num_threads 3 ' \
                  f'-max_target_seqs 10 -outfmt 5 -out "{self.XML}"'
            start = time()
            self.log.debug(arg)
            try:
                run(arg, shell=True, check=True)
                self.log.info(f'finished BLAST+ in {(time() - start):.2f} sec')
            except CalledProcessError as e:
                self.log.exception('BLAST failed, returned %d\n%s'
                                   % (e.returncode, e.output.decode('utf-8') if e.output is not None else ''))
                return None

            # read in metadata tsv
            self.df = pd.read_csv(self.TSV, sep='\t', dtype={'id': str})
            self.df.set_index(['id', 'gene'], inplace=True)
            self.df['BLAST_species'] = ''
            self.df['pid'] = nan
            self.df['extra_species'] = ''

            # parse .xml and write best hits info to metadata
            missing_seqs = list()
            for entry in SearchIO.parse(self.XML, 'blast-xml'):
                if not self._parse(entry):
                    # no hits found -> save seq_id for online NCBI BLAST
                    missing_seqs.append(entry.id)

            # write to TSV
            self.df.to_csv(self.TSV, sep='\t', na_rep='', header=True, index=True)
            self.log.debug('wrote updated metadata to %s' % self.TSV)

            return missing_seqs  # regular

        return []  # on thread kill

    def _blast_remote(self, missing_seqs):
        """
        Write seqs missing from local BLAST+ db to .FASTA and search in NCBI nucleotides.
        Updates non-seq data .csv again.
        :return:
        """
        while not self._stop_event.wait(.1):
            self.log.warning('online BLAST: %d seq%s missing from %s'
                             % (len(missing_seqs), 's' if len(missing_seqs) > 1 else '', self.db))

            # run remote NCBI BLAST via Biopython
            runs = ceil(len(missing_seqs) / 10)
            self.log.debug('BLASTing online in %d runs.' % runs)
            # prep file paths
            self.www_XML = [self.www_XML[:-4] + '_%d.xml' % run for run in range(1, runs + 1)]

            sleep_time = 11 if runs > 1 else 0
            for run in range(runs):
                start = time()
                # create query search string
                records = '\n'.join([self.seqdata[self.gene][seq_id].format('fasta')
                                     for seq_id in missing_seqs[run * 10: run * 10 + 10]])
                try:
                    # raise urllib.error.URLError('testing')
                    handle = NCBIWWW.qblast(program='blastn', database=self.remote_db,
                                            sequence=records, format_type='XML', hitlist_size=10)
                    # save as .xml
                    with open(self.www_XML[run], 'w') as fh:
                        fh.write(handle.read())
                    self.log.info('remote BLAST %d:%d complete after %.2f sec' % (run + 1, runs, time() - start))
                    self._parse_remote_result([self.www_XML[run]])
                    sleep(sleep_time)
                except urllib.error.URLError as offline:
                    self.log.warning('online BLAST aborted, no internet connection')
                    return
                except Exception as e:
                    self.log.exception(e)
                    sleep(sleep_time)

            self.log.info('finished remote BLAST')
            return

    def _parse(self, xml_entry):
        """
        Parse genus, species, pid, share of hits and other species from XML entry,
        assuming the calculated pid will only fall.
        :param xml_entry:
        :return: True if successful, False if no hits or Error.
        """
        if not xml_entry:
            return False
        try:
            # get row for selected gene
            row = self.df.loc[xml_entry.id, self.gene]
        except KeyError:
            try:
                rows = self.df.loc[xml_entry.id]
                self.gene = [rows.index[0], self.gene]
                row = rows.iloc[0]
            except KeyError:
                return False
        try:
            sp_counts, sp_pids = dict(), dict()
            last_pid = 0

            for hit in xml_entry:
                _def = hit.description.split(',')[0].strip().split(' ')
                genus, species = _def[0], _def[1]

                if species == 'cf.':
                    species += _def[2]
                    self.log.warning('found a cf. but that\'s fine')
                species = f'{genus} {species}'

                # parse % identity from high-scoring pair
                _hsp = hit.hsps[0]
                pid = _hsp.ident_num * 100 / (_hsp.aln_span + _hsp.gap_num)

                sp_pids[species] = pid
                sp_counts[species] = sp_counts.get(species, 0) + 1

                if pid != last_pid and len(set(sp_pids.values())) >= 2:
                    break  # as soon as there are two different pids, there must be two different species
                last_pid = pid

            finders = list()
            if len(sp_pids) == 2:
                # easiest case: found two species with two different pids
                four = [f for fs in [[k, v] for k, v in sp_pids.items()] for f in fs]
                if four[3] > four[1]:
                    finders.append([four.pop(3), four.pop(2)])
                while four:
                    finders.append([four.pop(1), four.pop(0)])
            else:
                # iterate over observed species along decreasing pid
                for best_pid in sorted(set(sp_pids.values()), reverse=True):
                    sp_best_pid = [sp for sp, pid in sp_pids.items() if pid == best_pid]
                    if len(sp_best_pid) == 1:
                        # best pid occurs exactly once
                        finders.append([best_pid, sp_best_pid[0]])
                    else:
                        # several species have best pid
                        sp_counts = {sp: count for sp, count in sp_counts.items() if sp in sp_best_pid}
                        most_freq = {sp for sp, c in sp_counts.items() if c == max(sp_counts.values())}
                        finders.append([best_pid, ' / '.join(most_freq)])
            if not finders:
                return False

            # write to dataframe
            for i, j in enumerate(['pid', 'BLAST_species']):
                row.at[j] = finders[0][i]
            if len(finders) > 1:
                row.at['extra_species'] = '%.2f: %s' % (finders[1][0], finders[1][1])
            if type(self.gene) == list:
                gene, self.gene = self.gene
                self.df.loc[xml_entry.id, gene] = row
            else:
                self.df.loc[xml_entry.id, self.gene] = row

            return True
        except Exception as ex:
            self.log.exception(ex)
            return False

    def _parse_remote_result(self, in_xmls):
        """
        Parses a list of XML files from previous online BLAST run
        :return:
        """
        # file check
        xmls = list()
        for xml in in_xmls:
            if path.isfile(xml):
                xmls.append(xml)
            else:
                self.log.warning('XML path invalid: %s' % xml)
        if not xmls:
            self.log.warning('no valid paths.')
            return
        # read in metadata tsv
        self.df = pd.read_csv(self.TSV, sep='\t', dtype={'id': str, 'BLAST_species': str})
        self.df.set_index(['id', 'gene'], inplace=True)
        if 'BLAST_species' not in self.df.columns:
            self.df['BLAST_species'] = ''
            self.df['pid'] = nan
            self.df['extra_species'] = ''

        errors = False
        # parse .xml and write best hits info to metadata
        with open(self.bad_seqs, 'a') as fh:
            for xml in xmls:
                for entry in SearchIO.parse(xml, 'blast-xml'):
                    if not self._parse(entry):
                        try:
                            if entry.id in self.df.index:
                                file, box = self.df.loc[entry.id, self.gene][['file', 'box']]
                                fh.write(f'{file}\t{entry.id}\t{box}\t{self.gene}\t'
                                         f'no hit in NCBI {self.remote_db} database\n')
                                self.log.error(f'{entry.id} no hit in NCBI {self.remote_db} db')
                            else:
                                self.log.error(f'no entry {entry.id} in metadata TSV')
                        except KeyError as ke:
                            errors = True
                            self.log.error(ke)
        if errors:
            self.log.warning('printing DataFrame for check')
            self.log.warning(self.df)

        # write to TSV
        self.df.to_csv(self.TSV, sep='\t', na_rep='', header=True, index=True)
        self.log.debug('wrote newly updated metadata to %s' % self.TSV)

        not_found = \
            set(self.df[self.df.BLAST_species == ''].index.get_level_values(0)) - \
            set(self.df[self.df.BLAST_species != ''].index.get_level_values(0))
        return not_found

    def _write_missing(self, not_found):
        """Save missing sequences to file"""
        if not not_found:
            self.log.info('all sequences found')
            return
        with open(self.missing_fasta, 'w') as fh:
            for seq_id in not_found:
                for gene in [self.gene] + list(self.seqdata.keys() - {self.gene}):
                    r = self.seqdata[gene].get(seq_id, False)
                    if r:
                        fh.write(r.format('fasta') + '\n')
                        break



#!/usr/bin/env python3
# 2021 Leo Kaindl

"""
Main module of the pipeline. Imports the other package modules and serves as the interface
between them. The `--visualize` and `--view` entry points are in the :class:`phylo` module.
"""

import sys

from ab12phylo_cmd import cli, i_o, msa, blast, ml, phylo


def _main():
    # get command line arguments
    args = cli.parser(sys.argv[1:]).args

    if not args.finish:
        if args.add_xml:
            print('starting -px BLAST XML run', file=sys.stderr)
            blaster = blast.blast_build(args, None)
            blaster.start()
            # reading an XML requires no waiting
            blaster.join()
            # visualize MSA for inspection
            phylo.mview_msa(args)
            print('finished -px BLAST XML run', file=sys.stderr)
            exit(0)

        if args.prepare:
            print('starting -p1 prep run', file=sys.stderr)

        # read in files
        reader = i_o.reader(args)

        # write .FASTAs and source information
        writer = i_o.writer(args, reader)

        # build & trim MSAs
        aligner = msa.msa_build(args, reader.seq_counts)

        # run BLAST and parse results de-synced
        blaster = blast.blast_build(args, reader)
        blaster.start()

        if args.prepare:
            # wait for BLAST if necessary
            blaster.join()
            # visualize MSA for inspection
            phylo.mview_msa(args)
            print('finished -p1 prep run', file=sys.stderr)
            exit(0)

        # build trees with raxml-ng or iqtree2
        ml_threads = ml.ml_build(args)
        ml_threads.run()

        # wait for BLAST if necessary
        blaster.join()

        # visualize best tree
        phylo.tree_build(args)

    else:
        print('starting -p2 finishing run', file=sys.stderr)
        # build trees with raxml-ng or iqtree2
        ml_threads = ml.ml_build(args)
        ml_threads.run()

        # visualize best tree
        phylo.tree_build(args)


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# 2021 Leo Kaindl

import logging
import os
import sys

LOG = logging.getLogger(__name__)

try:
    import gi

    gi.require_version('Gtk', '3.0')
except ModuleNotFoundError as ex:
    LOG.exception(ex)
    from subprocess import run, PIPE
    import shlex
    import shutil
    import getpass

    yes = {'y', 'yes'}
    yes_or_no = yes.union({'n', 'no'})

    # anaconda installation
    if os.path.exists(os.path.join(sys.prefix, "conda-meta")):
        run_0 = f'conda activate {sys.prefix.split("/")[-1]}'
        run_1 = 'conda install -y -c conda-forge pygobject gtk3 '
        run_2 = 'conda install -y -c conda-forge adwaita-icon-theme hicolor-icon-theme '
        print(f'\nPyGObject was not found on your python at '
              f'{sys.prefix}. It looks like you are using anaconda, '
              f'so the missing dependencies can be installed via:\n'
              f'\033[94m{run_0}\n{run_1}\n{run_2}\033[0m\n')

        if sys.prefix.endswith('base'):
            print(f'\033[91mInstalling to the \033[0mbase\033[91m environment '
                  f'is likely to cause package conflicts!\033[0m\nYou can find '
                  f'tested environments for Linux https://raw.githubusercontent'
                  f'.com/lkndl/ab12phylo/main/recipe/ab1.yaml and Windows https'
                  f'://raw.githubusercontent.com/lkndl/ab12phylo/main/recipe/wi'
                  f'n.yaml in the AB12PHYLO repo. Install one with:\n'
                  f'\033[94mconda env create -f ab1.yaml\n'
                  f'conda activate ab1\033[0m\n')

        answer = ''
        while answer not in yes_or_no:
            answer = input(f'Install now? [y/n]').lower().strip()
        if answer in {'n', 'no'}:
            exit(1)

        print(f'\033[94m{sys.prefix}\033[0m')
        for arg in [run_1, run_2]:
            print(f'\033[94m{arg}\033[0m')
            proc = run(shlex.split(arg))
            if proc.returncode != 0:
                exit(1)
        try:
            import gi

            gi.require_version('Gtk', '3.0')
        except Exception as ex:
            LOG.exception(ex)
    else:
        # non-anaconda case
        print(f'\nPyGObject was not found on your python at '
              f'{sys.prefix}.')
        if shutil.which('apt') is None:
            print('On this system, please install AB12PHYLO to a python3 '
                  'environment managed by the anaconda package manager.')
            exit(1)

        apt_call = 'sudo apt install python3-gi python3-gi-cairo gir1.2-gtk-3.0'
        print(f'It looks like you are using a Linux distro '
              f'that uses the Debian package manager APT. '
              f'GTK3 is often already installed on such a system, '
              f'but you can try installing the missing packages '
              f'with\n\033[94m{apt_call}\033[0m\nbut this might break '
              f'your system!\nA safer method would be using the '
              f'anaconda package manager and setting up a tested '
              f'environment: https://raw.githubusercontent.com/lkndl/'
              f'ab12phylo/main/recipe/ab1.yaml. Create it via:\n'
              f'\033[94mconda env create -f ab1.yaml\n'
              f'conda activate ab1\033[0m\n')
        answer = ''
        while answer not in yes_or_no:
            answer = input(f'Try apt installation now? \033[91mUse with '
                           f'caution!\033[0m [y/n]').lower().strip()
        if answer in {'n', 'no'}:
            exit(1)

        print(f'\033[94m{apt_call}\033[0m')
        try:
            proc = run(shlex.split(apt_call))
            if proc.returncode != 0:
                exit(1)
        except Exception as ex:
            LOG.exception(ex)

from ab12phylo import repo
from ab12phylo.gtk_blast import blast_page
from ab12phylo.gtk_gbl import gbl_page
from ab12phylo.gtk_io import io_page
from ab12phylo.gtk_ml import ml_page
from ab12phylo.gtk_msa import msa_page
from ab12phylo.gtk_qal import qal_page
from ab12phylo.gtk_rgx import rgx_page
from ab12phylo.gtk_tree import tree_page


# set the icon theme
# Gtk.Settings.get_default().set_property('gtk-icon-theme-name', 'Papirus-Dark-Maia')
# Gtk.Settings.get_default().set_property('gtk-theme-name', 'Matcha-dark-sea')


class ab12phylo_app(io_page, rgx_page, qal_page, msa_page,
                    gbl_page, blast_page, ml_page, tree_page):

    def __init__(self):
        super().__init__()
        self.supers = [io_page, rgx_page, qal_page, msa_page,
                       gbl_page, blast_page, ml_page, tree_page]
        self.re_runs = {2: self.start_trim,  # 1: self.start_read,
                        4: self.start_gbl, 7: self.start_phy}

    def load(self, path, *args):
        super().load(path)
        for module in self.supers:
            try:
                module.reload_ui_state(self)
            except AttributeError as e:
                LOG.debug(e)

    def reload_ui_state(self, page=False, *args):
        if not page:
            return
        try:
            self.supers[page].reload_ui_state(self)
        except AttributeError as e:
            LOG.debug(e)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        # tell the MSA pre-set about it
        if 'aligner' in self.iface:
            self.refresh_paths()

    def test(self, *args):
        super().test(self, args)
        self.supers[1].refresh(self)

    def refresh(self, page=-1, *args):
        """
        Call the refresh function of the current page and hide or show
        the appropriate buttons.
        """
        iface = self.iface
        if page == -1 or type(page) != int:
            page = iface.notebook.get_current_page()
        self.supers[page].refresh(self)
        # hide or show these two actions depending on applicability
        iface.refresh.props.visible = bool(page in self.re_runs)
        iface.reset.props.visible = page in {1, 2, 4, 7}
        iface.gene_roll.props.visible = page == 2
        iface.next.props.visible = page != 7
        iface.back.props.visible = page != 0
        iface.open_test.props.visible = page == 0
        iface.helper.set_markup(repo.help.get(page, ''))

    def re_run(self, *args):
        """
        Depending on the currently visible page, re-run the matching background task.
        Handles the Refresh button.
        """
        page = self.iface.notebook.get_current_page()
        if page in self.re_runs:
            self.re_runs[page]()
        else:
            raise RuntimeWarning(f'page {page} has no re-run')

    def reset(self, *args):
        """
        Depending on the currently visible page (either rgx, trim, gbl or tree), reset the
        regex table or the tree. Handles the Reset button.
        """
        page = self.iface.notebook.get_current_page()
        if page == 1:
            self.set_changed(1)
            self.reset_columns(do_parse=True)
        elif page == 2:
            self.reset_x_scale()
            self.data.qal.ignore_ids = {g: set() for g in self.data.genes}
            self.start_read(run_after=[self.start_trim])
            # self.start_trim()
        elif page == 4:
            self.reset_x_scale()
            self.undrop_seqs()
        elif page == 7:
            self.reset_tree()

    def proceed(self, widget=None, page=None):
        """
        The function connected to the _Next button. For pages with a background thread,
        this will start it and instruct it to re-run this function afterwards; to make the
        application proceed only upon thread completion.
        :param widget: optional, for use as callback
        :param gui:
        :param page:
        :return:
        """
        data = self.data
        iface = self.iface
        page = iface.notebook.get_current_page() if not page else page

        if self.get_errors(page):
            self.show_notification('There are still errors on the page!')
            return

        if self.get_changed(page):
            if page == 0:
                self.supers[page].refresh(self)
                self.reset_columns(do_parse=True)
            elif page == 1:
                self.supers[page].refresh(self)
                if 1 < sum(data.rx_fired) < 5:
                    self.show_notification('Make sure all columns have been parsed.')
                    return
                self.start_read(run_after=[self.init_gene_roll, self.proceed])
                return  # leave this alone
            elif page == 2:
                self.trim_all(run_after=[self.proceed])
                return
            elif page == 3:
                self.start_align(None, run_after=[self.proceed])
                return
            elif page == 4:
                self.start_gbl()
            elif page == 5:
                iface.tempspace.df.to_csv(
                    self.wd / repo.PATHS.tsv, sep='\t', na_rep='', header=True, index=True)
                self.set_changed(5, False)
            elif page == 6:
                self.start_phy()
            self.set_changed(page, False)

        # then proceed
        iface.notebook.next_page()
        data.page = iface.notebook.get_current_page()
        LOG.debug('proceeded to page %d' % data.page)

    def tree_modify(self, action, sel_idx=None):
        """
        Fetch the names of the selected nodes.
        :param action:
        :param sel_idx:
        :return: {action: [matching names]} that will be added to the existing modifications
        """
        if not sel_idx:
            sel_idx = [tp.get_indices()[0] for tp in self.iface.tree_sel.get_selected_rows()[1]]
        if not sel_idx:
            self.show_notification('Cannot %s, no selection' % action.get_name(), secs=2)
            return
        if not self.iface.tree:
            self.show_notification('No tree in memory, re-drawing first', secs=2)
            self.start_phy(run_after=(self.tree_modify, (action, sel_idx)))
            return

        # return a list of tip labels
        idsp = [(_id, sp) for i, (_id, sp) in enumerate(
            self.data.tree_anno_model.get_column((0, 1))) if i in sel_idx]

        names = list()
        for _id, sp in idsp:
            if sp.startswith('REF_'):
                names.append(sp.split(':')[0])
            else:
                names.append(_id.replace(', ', '~'))
        LOG.debug('re-plot and %s' % action.get_name())
        self.start_phy({action.get_name(): names})

    def delete_and_ignore_rows(self, *args):
        if super().delete_and_ignore_rows(*args):
            page = args[2]
            if page == 2:
                self.iface.view_qal.grab_focus()
                self.start_trim()
            elif page == 4:
                self.iface.view_gbl.grab_focus()
                self.drop_seqs()
        return True


def main():
    app = ab12phylo_app()
    exit_status = app.run(sys.argv)
    sys.exit(exit_status)


if __name__ == '__main__':
    main()
